import os

ATASKQ_CONFIG = os.getenv("ATASKQ_CONFIG")
# for dev purposes add additional config for server defaulted to normal config
ATASKQ_SERVER_CONFIG = os.getenv("ATASKQ_SERVER_CONFIG", ATASKQ_CONFIG)



import json

import pytest

from . import TaskQ, Task, targs
from .tasks_utils.write_to_file_tasks import write_to_file, write_to_file_mp_lock
from .__main__ import main


def test_empty():
    try:
        args = []
        main(args=args)
    except Exception as ex:
        pytest.fail("main without args failed")


def test_run_job(tmp_path, config):
    filepath = tmp_path / "file.txt"
    with open(configpath := tmp_path / "config.json", "w") as f:
        json.dump(config, f)

    taskq = TaskQ(config=config).create_job()

    taskq.add_tasks(
        [
            Task(entrypoint=write_to_file, targs=targs(filepath, "task 0\n")),
            Task(entrypoint=write_to_file, targs=targs(filepath, "task 1\n")),
            Task(entrypoint=write_to_file, targs=targs(filepath, "task 2\n")),
        ]
    )

    args = ["run", "-c", str(configpath), "--job-id", str(taskq.job_id)]
    main(args=args)

    assert filepath.exists()
    assert filepath.read_text() == "task 0\n" "task 1\n" "task 2\n"


from . import TaskQ, Task, targs
from .tasks_utils.write_to_file_tasks import write_to_file
from .__main__ import main


def test_run_level(tmp_path, config):
    filepath = tmp_path / "file.txt"
    with open(configpath := tmp_path / "config.json", "w") as f:
        json.dump(config, f)

    taskq = TaskQ(config=config).create_job()

    taskq.add_tasks(
        [
            Task(entrypoint=write_to_file, level=0, targs=targs(filepath, "task 0\n")),
            Task(entrypoint=write_to_file, level=0, targs=targs(filepath, "task 1\n")),
            Task(entrypoint=write_to_file, level=1, targs=targs(filepath, "task 2\n")),
            Task(entrypoint=write_to_file, level=2, targs=targs(filepath, "task 3\n")),
            Task(entrypoint=write_to_file, level=3, targs=targs(filepath, "task 4\n")),
        ]
    )

    # run level 0
    args = ["run", "-c", str(configpath), "--job-id", str(taskq.job_id), "--level", "0"]
    main(args=args)

    assert filepath.exists()
    assert filepath.read_text() == "task 0\n" "task 1\n"

    # run level 1
    args = ["run", "-c", str(configpath), "--job-id", str(taskq.job_id), "--level", "1"]
    main(args=args)

    assert filepath.exists()
    assert filepath.read_text() == "task 0\n" "task 1\n" "task 2\n"

    # run level 2, 3
    args = ["run", "-c", str(configpath), "--job-id", str(taskq.job_id), "--level", "2", "4"]
    main(args=args)

    assert filepath.exists()
    assert filepath.read_text() == "task 0\n" "task 1\n" "task 2\n" "task 3\n" "task 4\n"


def test_run_concurrency(tmp_path, config):
    filepath = tmp_path / "file.txt"
    with open(configpath := tmp_path / "config.json", "w") as f:
        json.dump(config, f)

    taskq = TaskQ(config=config).create_job()

    taskq.add_tasks(
        [
            Task(entrypoint=write_to_file_mp_lock, targs=targs(filepath, "@{pid}\n", sleep=1)),
            Task(entrypoint=write_to_file_mp_lock, targs=targs(filepath, "@{pid}\n", sleep=1)),
            Task(entrypoint=write_to_file_mp_lock, targs=targs(filepath, "@{pid}\n", sleep=1)),
            Task(entrypoint=write_to_file_mp_lock, targs=targs(filepath, "@{pid}\n", sleep=1)),
            Task(entrypoint=write_to_file_mp_lock, targs=targs(filepath, "@{pid}\n", sleep=1)),
        ]
    )

    args = ["run", "-c", str(configpath), "--job-id", str(taskq.job_id), "--concurrency", "3"]
    main(args=args)

    assert filepath.exists()
    assert len(set([l for l in filepath.read_text().split("\n") if l])) == 3



import pytest


@pytest.mark.parametrize("conn_type", ["sqlite", "pg", "http"])
def test_conn_type_check(conn_type, config):
    conn = config["connection"]
    if f"{conn_type}://" not in conn:
        pytest.skip()



import starlette.datastructures


async def form_data_array(data: dict):
    ret = []
    for k, v in data.items():
        import pickle

        if isinstance(v, starlette.datastructures.UploadFile):
            v = await v.read()
        # expect format of index.key
        assert "." in k
        i, *k = k.split(".")
        i = int(i)
        k = ".".join(k)
        # expect monotonic rising items
        if i == len(ret):
            ret += [dict()]
        elif i == len(ret) - 1:
            pass
        else:
            # should never get here
            raise RuntimeError(f"unexpected item index '{i}'. len res: {len(ret)}")
        ret[i][k] = v

    return ret



import asyncio
import logging

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from contextlib import asynccontextmanager

from ataskq.handler import DBHandler, from_config
from ataskq.env import ATASKQ_SERVER_CONFIG

logger = logging.getLogger("uvicorn")


def db_handler() -> DBHandler:
    return from_config(ATASKQ_SERVER_CONFIG or "server")


async def set_timeout_tasks_task():
    dbh = db_handler()
    while True:
        logger.info(f"Set Timeout Tasks - {dbh.config['background']['pulse_timeout_interval']} sec interval")
        dbh.fail_pulse_timeout_tasks(dbh.config["monitor"]["pulse_timeout"])
        await asyncio.sleep(dbh.config["background"]["pulse_timeout_interval"])


@asynccontextmanager
async def lifespan(app: FastAPI):
    logger.info("enter lifspan")

    logger.info("init db")
    db_handler().init_db()

    task = asyncio.create_task(set_timeout_tasks_task())

    # Load the ML model
    yield
    # Clean up the ML models and release the resources
    logger.info("cancel task")
    task.cancel()
    logger.info("exit lifspan")


app = FastAPI(lifespan=lifespan)

# allow all cors
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.get("/")
@app.get("/health")
async def health():
    return "Background manager is running"



from abc import ABC, abstractmethod
from typing import Union, List, Dict, Callable


class IModelSerializer(ABC):
    @staticmethod
    @abstractmethod
    def i2m_serialize() -> Dict[type, Callable]:
        pass

    @staticmethod
    @abstractmethod
    def m2i_serialize() -> Dict[type, Callable]:
        pass


class IModel(ABC):
    @staticmethod
    @abstractmethod
    def id_key():
        pass

    @staticmethod
    @abstractmethod
    def table_key():
        pass

    @staticmethod
    def children():
        return dict()

    @staticmethod
    @abstractmethod
    def i2m(cls, kwargs: Union[dict, List[dict]], serializer: IModelSerializer) -> Union[dict, List[dict]]:
        pass

    @staticmethod
    @abstractmethod
    def from_interface(cls, kwargs: Union[dict, List[dict]], serializer: IModelSerializer):
        pass

    @staticmethod
    @abstractmethod
    def m2i(cls, kwargs: Union[dict, List[dict]], serializer: IModelSerializer) -> Union[dict, List[dict]]:
        pass

    @abstractmethod
    def to_interface(self, serializer: IModelSerializer) -> dict:
        pass



import copy
import os

import pytest

from .config import load_config, get_config_set
from .config.config import CONFIG_FORMAT, DEFAULT_CONFIG


def assert_instance(ref, format):
    if ref is None:
        return True
    elif isinstance(ref, (int, float)) and issubclass(format, (int, float)):
        return True
    else:
        return isinstance(ref, format)


def assert_config(ref_config: dict, test_config: dict, format=CONFIG_FORMAT, path="", count=None):
    if count is None:
        count = [0]

    for k in ref_config.keys():
        kpath = (path and path + ".") + k
        assert k in ref_config, f"{kpath} missing in dst_config"
        if isinstance(ref_config[k], dict):
            assert_config(ref_config[k], test_config[k], format=format[k], path=kpath, count=count)
        else:
            # overwrite config with environment value
            assert assert_instance(ref_config[k], format[k]), f"config '{kpath}' type mismatch"
            assert ref_config[k] == test_config[k], f"config '{kpath}' value mismatch"
            count[0] += 1

    return count[0]


def test_load_default_none():
    config = load_config(environ=False)
    count = assert_config(get_config_set(), config)
    # sanity
    assert count == 12, "invalid number of configurations."


def test_load_default():
    config = load_config(DEFAULT_CONFIG, environ=False)
    assert_config(get_config_set(), config)


def test_load_client_preset():
    config = load_config("client", environ=False)

    ref = copy.deepcopy(get_config_set())
    ref["connection"] = "http://localhost:8080"
    ref["handler"]["db_init"] = False
    ref["run"]["fail_pulse_timeout"] = False

    assert_config(ref, config)


def test_load_custom():
    config = load_config({"connection": "test", "run": {"wait_timeout": 100}}, environ=False)

    ref = copy.deepcopy(get_config_set())
    ref["connection"] = "test"
    ref["run"]["wait_timeout"] = 100.0

    assert_config(ref, config)


def test_load_custom2():
    config = load_config([{"connection": "test", "run": {"wait_timeout": 100}}], environ=False)

    ref = copy.deepcopy(get_config_set())
    ref["connection"] = "test"
    ref["run"]["wait_timeout"] = 100.0

    assert_config(ref, config)


def test_load_custom_and_preset():
    config = load_config(["client", {"connection": "test"}], environ=False)

    ref = copy.deepcopy(get_config_set())
    ref["connection"] = "test"
    ref["handler"]["db_init"] = False
    ref["run"]["fail_pulse_timeout"] = False

    assert_config(ref, config)


def test_invalid_config_format():
    with pytest.raises(ValueError) as excinfo:
        load_config({"run": {"wait_timeout": "asd"}}, environ=False)

    assert "config 'run.wait_timeout' value 'asd' can't be cast to 'float'" == str(excinfo.value)


def test_invalid_config_type():
    with pytest.raises(RuntimeError) as excinfo:
        load_config(config=2)

    assert "invalid config type. supported types: ['str', 'Path', 'dict']" == str(excinfo.value)


def test_invalid_config_type_in_list():
    with pytest.raises(RuntimeError) as excinfo:
        load_config(config=[DEFAULT_CONFIG, 2])

    assert "onvalid config[1] element type. supported types: ['str', 'Path', 'dict']" == str(excinfo.value)


def test_loaded_config():
    c = {"a": 3, "_loaded": True}
    config = load_config(c)

    assert id(c) == id(config)



"""Special tests scripts which tests environment variables handling
   outside test suite scope (due to conflicts)
"""

import os
import sys

sys.path.append(f"{os.path.dirname(__file__)}/..")
from ataskq.config import load_config
from ataskq.config.config import DEFAULT_CONFIG


def load_with_env_override():
    os.environ["ataskq.connection"] = "test"
    os.environ["ataskq.run.wait_timeout"] = "111"
    config = load_config(DEFAULT_CONFIG)

    assert config["connection"] == "test"
    assert config["run"]["wait_timeout"] == 111.0

    os.environ.pop("ataskq.connection")
    os.environ.pop("ataskq.run.wait_timeout")


def load_with_env_override2():
    os.environ["ataskq_connection"] = "test-2"
    os.environ["ataskq_run_wait_timeout"] = "111.3"
    config = load_config(DEFAULT_CONFIG)

    assert config["connection"] == "test-2"
    assert config["run"]["wait_timeout"] == 111.3

    os.environ.pop("ataskq_connection")
    os.environ.pop("ataskq_run_wait_timeout")


if __name__ == "__main__":
    load_with_env_override()
    load_with_env_override2()



import logging
import asyncio
from pathlib import Path

from fastapi import FastAPI, Request, Depends
from fastapi.responses import FileResponse, RedirectResponse
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles


from ataskq.handler import DBHandler, from_config
from ataskq.handler.rest_handler import RESTHandler as rh
from ataskq.models import Model, __MODELS__
from ataskq.env import ATASKQ_SERVER_CONFIG

# from .form_utils import form_data_array

logger = logging.getLogger("uvicorn")


# DB Handler
def db_handler() -> DBHandler:
    return from_config(ATASKQ_SERVER_CONFIG or "server")


app = FastAPI()


# allow all cors
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# Custom exception handler


# async def custom_exception_handler(request: Request, exc: Exception):
#     raise exc

# # Adding the custom exception handler to the app
# app.add_exception_handler(Exception, custom_exception_handler)

# Example route with intentional exception


# static folder
app.mount("/www", StaticFiles(directory=Path(__file__).parent / "www"), name="www")


@app.get("/")
async def root():
    # return "Welcome to A-TASK-Q Server"
    return RedirectResponse("/custom_query/jobs_status")


@app.get("/health")
async def health():
    return "Server is running"


@app.get("/api")
async def api():
    return {"message": "Welcome to A-TASK-Q Server API"}


####################
# Custom Query API #
####################
@app.get("/api/custom_query/take_next_task")
async def take_next_task(
    request: Request,
    dbh: DBHandler = Depends(db_handler),
):
    # take next task
    action, task = dbh.take_next_task(**request.query_params)
    task = rh.to_interface(task) if task is not None else None

    return dict(action=action, task=task)


@app.get("/api/custom_query/jobs_status")
async def jobs_status(request: Request, dbh: DBHandler = Depends(db_handler)):
    ret = dbh.jobs_status(**request.query_params)

    return ret


@app.get("/api/custom_query/tasks_status")
async def tasks_status(request: Request, dbh: DBHandler = Depends(db_handler)):
    ret = dbh.tasks_status(**request.query_params)

    return ret


#############
# Model API #
#############
@app.get("/api/{model}")
async def get_model_all(model: str, request: Request, dbh: DBHandler = Depends(db_handler)):
    # logger.info(f"query_params: {request.query_params}")
    model_cls = __MODELS__[model]
    mkwargs = model_cls.get_all_dict(_handler=dbh, **request.query_params)
    ikwargs = rh.m2i(model_cls, mkwargs)

    return ikwargs


@app.get("/api/{model}/count")
async def count_model_all(model: str, request: Request, dbh: DBHandler = Depends(db_handler)):
    model_cls = __MODELS__[model]
    count = model_cls.count_all(_handler=dbh, **request.query_params)

    return count


@app.get("/api/{model}/{model_id}")
async def get_model(model: str, model_id: int, dbh: DBHandler = Depends(db_handler)):
    model_cls = __MODELS__[model]
    mkwargs = model_cls.get_dict(model_id, _handler=dbh)
    ikwargs = rh.m2i(model_cls, mkwargs)

    return ikwargs


@app.post("/api/{model}")
async def create_model(model: str, request: Request, dbh: DBHandler = Depends(db_handler)):
    model_cls: Model = __MODELS__[model]
    ikwargs = await request.json()
    mkwargs = rh.i2m(model_cls, ikwargs)
    model_id = dbh.create(model_cls, **mkwargs)

    return model_id


@app.post("/api/{model}/bulk")
async def create_model_bulk(model: str, request: Request, dbh: DBHandler = Depends(db_handler)):
    model_cls: Model = __MODELS__[model]
    ikwargs = await request.json()
    mkwargs = rh.i2m(model_cls, ikwargs)
    model_ids = dbh.create_bulk(model_cls, mkwargs)

    return model_ids


@app.put("/api/{model}/{model_id}")
async def update_model(model: str, model_id: int, request: Request, dbh: DBHandler = Depends(db_handler)):
    model_cls: Model = __MODELS__[model]
    ikwargs = await request.json()
    mkwargs = rh.i2m(model_cls, ikwargs)
    dbh.update(model_cls, model_id, **mkwargs)

    return {model_cls.id_key(): model_id}


@app.delete("/api/{model}/{model_id}")
async def delete_model(model: str, model_id: int, dbh: DBHandler = Depends(db_handler)):
    model_cls: Model = __MODELS__[model]
    dbh.delete(model_cls, model_id)

    return {model_cls.id_key(): model_id}


#######
# WWW #
#######
@app.get("/db/{model}")
async def show_db(model: str):
    return FileResponse(Path(__file__).parent / "www" / "index.html")


@app.get("/custom_query/{query}")
async def show_custom_qeury(query: str):
    return FileResponse(Path(__file__).parent / "www" / "index.html")


@app.get("/custom_query/{query}/{job_id}")
async def show_job_custom_query(query: str, job_id: int):
    return FileResponse(Path(__file__).parent / "www" / "index.html")



import re
from typing import NamedTuple, List
import sqlite3
from datetime import datetime

from ..imodel import IModel
from .handler import to_datetime, from_datetime
from .db_handler import DBHandler, transaction_decorator


class SqliteConnection(NamedTuple):
    path: str

    def __str__(self):
        return f"sqlite://{self.path}"


class SQLite3DBHandler(DBHandler):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    @staticmethod
    def from_connection_str(conn):
        format = "sqlite://path"
        pattern = r"sqlite://(?P<path>.+)$"
        match = re.match(pattern, conn)

        if not match:
            raise Exception(f"conn must be in '{format}', ex: 'sqlite://ataskq.db.sqlite3'")

        path = match.group("path")
        ret = SqliteConnection(path=path)

        return ret

    @staticmethod
    def m2i_serialize():
        type_handlers = {
            datetime: lambda v: from_datetime(v),
        }

        return type_handlers

    @staticmethod
    def i2m_serialize():
        type_handlers = {
            datetime: lambda v: to_datetime(v),
        }

        return type_handlers

    @property
    def pragma_foreign_keys_on(self):
        return "PRAGMA foreign_keys = ON"

    @property
    def format_symbol(self):
        return "?"

    @property
    def connection(self):
        return self._connection

    @property
    def db_path(self):
        return self._connection.path

    @property
    def bytes_type(self):
        return "MEDIUMBLOB"

    @property
    def primary_key(self):
        return "INTEGER PRIMARY KEY AUTOINCREMENT"

    @property
    def timestamp_type(self):
        return "DATETIME"

    def timestamp(self, ts):
        return f"'{ts}'"

    @property
    def begin_exclusive(self):
        return "BEGIN EXCLUSIVE"

    @property
    def for_update(self):
        return ""

    def connect(self):
        conn = sqlite3.connect(self.db_path)
        return conn

    @transaction_decorator()
    def _create_bulk(self, c, model_cls: IModel, ikwargs: List[dict]) -> List[int]:
        # todo: consolidate all ikwargs with same keys to single insert command
        model_ids = []
        for v in ikwargs:
            d = {k: v for k, v in v.items() if model_cls.id_key() not in k}
            keys = list(d.keys())
            values = list(d.values())
            c.execute(
                f'INSERT INTO {model_cls.table_key()} ({", ".join(keys)}) VALUES ({", ".join([self.format_symbol] * len(keys))})',
                values,
            )
            model_id = c.lastrowid
            model_ids.append(model_id)

        return model_ids



try:
    from .version import __version__, __build__
except Exception:
    __version__ = "0.0.0"
    __build__ = "dev"

__schema_version__ = 5

from .taskq import TaskQ, targs
from .models import Job, Task, EStatus



import pytest

from .models import Model, __MODELS__, Job
from .handler import Handler, from_config, register_handler, unregister_handler


@pytest.fixture(scope="function")
def handler(config):
    handler = from_config(config)
    register_handler("test_handler", handler)
    yield handler
    unregister_handler("test_handler")


@pytest.fixture
def jhandler(handler) -> Handler:
    return handler.create_job()


def init(model_cls, **kwargs):
    # todo: better handle not Null fields (take from schema in future)
    annotations = model_cls.__annotations__.keys()
    if "entrypoint" in annotations and "entrypoint" not in kwargs:
        kwargs["entrypoint"] = "dummy entry point"
    if "job_id" in annotations and "job_id" not in kwargs and not issubclass(model_cls, Job):
        job = create(Job)
        kwargs["job_id"] = job.job_id

    ret = model_cls(**kwargs)

    return ret


def create(model_cls, **kwargs):
    m = init(model_cls, **kwargs)
    m = m.create()

    return m


@pytest.mark.parametrize("model_cls", __MODELS__.values(), ids=__MODELS__.keys())
def test_create(handler, model_cls):
    m = create(model_cls, name="test name")

    count = len(model_cls.get_all())
    assert count == 1

    assert m.name == "test name"


def assert_model(m_src, m_rec, model_cls, first_id, i):
    assert isinstance(m_rec, model_cls), f"index: '{i}'"
    assert m_src.name == m_rec.name == f"test {i+1}", f"index: '{i}'"
    assert m_src.__dict__ == m_rec.__dict__, f"index: '{i}'"
    assert getattr(m_src, m_src.id_key()) == i + first_id, f"index: '{i}'"
    assert getattr(m_rec, m_rec.id_key()) == i + first_id, f"index: '{i}'"


@pytest.mark.parametrize("model_cls", __MODELS__.values(), ids=__MODELS__.keys())
def test_get_all(handler, model_cls):
    m1 = create(model_cls, name="test 1")
    m2 = create(model_cls, name="test 2")
    m3 = create(model_cls, name="test 3")
    data = [m1, m2, m3]
    m_all = model_cls.get_all()
    assert len(m_all) == 3

    first_id = getattr(m1, m1.id_key())
    for i in range(len(data)):
        m_src = data[i]
        m_rec = m_all[i]
        assert_model(m_src, m_rec, model_cls, first_id, i)


@pytest.mark.parametrize("model_cls", __MODELS__.values(), ids=__MODELS__.keys())
def test_get_all_where(handler, model_cls: Model):
    m1 = create(model_cls, name="test 1")
    m2 = create(model_cls, name="test 2")
    m3 = create(model_cls, name="test 3")

    m_all = model_cls.get_all(_where="name='test 1'")
    assert len(m_all) == 1

    m_rec = m_all[0]
    first_id = getattr(m1, m1.id_key())
    assert_model(m1, m_rec, model_cls, first_id, 0)


@pytest.mark.parametrize("model_cls", __MODELS__.values(), ids=__MODELS__.keys())
def test_get(handler, model_cls):
    m1 = create(model_cls, name="test 1")
    m2 = create(model_cls, name="test 2")
    m3 = create(model_cls, name="test 3")
    data = [m1, m2, m3]

    first_id = getattr(m1, m1.id_key())
    m_all = [model_cls.get(i + first_id) for i in range(len(data))]
    for i in range(len(data)):
        m = data[i]
        m_rec = m_all[i]
        assert isinstance(m_rec, model_cls), f"index: '{i}'"
        assert m.name == m_rec.name == f"test {i+1}", f"index: '{i}'"
        assert m.__dict__ == m_rec.__dict__, f"index: '{i}'"
        assert getattr(m, m.id_key()) == i + first_id, f"index: '{i}'"
        assert getattr(m_rec, m_rec.id_key()) == i + first_id, f"index: '{i}'"


@pytest.mark.parametrize("model_cls", __MODELS__.values(), ids=__MODELS__.keys())
def test_delete(handler, model_cls):
    m = create(model_cls)

    count = len(model_cls.get_all())
    assert count == 1

    m.delete()

    count = len(model_cls.get_all())
    assert count == 0


MODEL_CHILD = [
    (model_cls, child_cls, parent_key)
    for model_cls in __MODELS__.values()
    for child_cls, parent_key in model_cls.children().items()
]


@pytest.mark.parametrize("model_cls, child_cls, parent_key", MODEL_CHILD)
def test_add_get_children(handler, model_cls: Model, child_cls, parent_key):
    m1 = create(model_cls)
    children = [init(child_cls, **{parent_key: None}) for i in range(3)]
    m1.add_children(child_cls, children)

    m2 = create(model_cls)
    children = [init(child_cls, **{parent_key: None}) for i in range(4)]
    m2.add_children(child_cls, children)

    assert getattr(m1, model_cls.id_key()) != getattr(m2, model_cls.id_key())

    rec_children = m1.get_children(child_cls)
    assert len(rec_children) == 3
    assert all([getattr(c, parent_key) == getattr(m1, model_cls.id_key()) for c in rec_children])

    rec_children = m2.get_children(child_cls)
    assert len(rec_children) == 4
    assert all([getattr(c, parent_key) == getattr(m2, model_cls.id_key()) for c in rec_children])



import argparse
import logging

from ataskq import TaskQ
from ataskq.config.config import CONFIG_SETS, DEFAULT_CONFIG


def init_logger(level=logging.INFO):
    logger = logging.getLogger("ataskq")

    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter(
            "%(asctime)s.%(msecs)03d [%(process)d] [%(levelname)s] %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
    )
    handler.setLevel(level)

    logger.addHandler(handler)
    logger.setLevel(level)

    return logger


def parse_number(number: str):
    try:
        ret = int(number)
        return ret
    except ValueError:
        pass

    try:
        ret = float(number)
        return ret
    except ValueError:
        raise ValueError(f"Failed to parse '{number}'")


def main(args=None):
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, description="ataskq command line interface help"
    )
    subparsers = parser.add_subparsers(title="commands", dest="command")
    run_p = subparsers.add_parser("run")

    run_p.add_argument(
        "--config",
        "-c",
        help=f"config preset {list(CONFIG_SETS.keys())} or path to file",
        default=DEFAULT_CONFIG,
    )
    run_p.add_argument("--job-id", "-jid", type=int, help="job id to run")
    run_p.add_argument("--level", "-l", type=int, nargs="+", help="job level to run")
    run_p.add_argument(
        "--concurrency", "-cn", type=parse_number, help="number of task execution processes to run in parallel"
    )

    args = parser.parse_args(args=args)

    # specific args handling
    if args.command == "run":
        if args.level is not None and len(args.level) == 1:
            args.level = args.level[0]
        init_logger()
        TaskQ(config=args.config, job_id=args.job_id).run(level=args.level, concurrency=args.concurrency)


if __name__ == "__main__":
    main()



import time


def hello_world():
    print("hello world")


def dummy_args_task(*args, **kwargs):
    if "sleep" in kwargs:
        time.sleep(kwargs["sleep"])
    print(f"entrypoint_with_args args: {args}, kwargs: {kwargs}")


def exception_task(etype=Exception, message="This is an exception task"):
    raise etype(message)



import copy
from typing import Union, List, get_args
from pathlib import Path
import json
import os

from .config import CONFIG_FORMAT, CONFIG_SETS, DEFAULT_CONFIG

CONFIG_TYPE = Union[str, Path, dict]


def _get_args_names(union):
    return [u.__name__ for u in get_args(union)]


def _load_config(config: List[dict], default: dict, format=CONFIG_FORMAT, path="", environ=False):
    """load config dict,

    Args:
        config (dict): _description_
        path (str, optional): _description_. Defaults to "".
    """
    ret = {}
    for k in format.keys():
        kpath = (path and path + ".") + k

        # recursive iterate dicts
        if isinstance(format[k], dict):
            ret[k] = _load_config(
                [c[k] for c in config if k in c],
                default=default[k],
                format=format[k],
                path=kpath,
                environ=environ,
            )
            continue

        # set default value
        ret[k] = default[k]

        # overwrite config with environment value (if exists)
        if environ:
            env_var = f"ataskq.{kpath}"
            if env_var in os.environ:
                try:
                    ret[k] = format[k](os.environ[env_var])
                    continue
                except:
                    raise ValueError(
                        f"Failed parsing config env variable '{env_var}' value '{os.environ[env_var]}' to '{format[k].__name__}'"
                    )

            env_var = env_var.replace(".", "_")
            if env_var in os.environ:
                try:
                    ret[k] = format[k](os.environ[env_var])
                    continue
                except:
                    raise ValueError(
                        f"Failed parsing config env variable '{env_var}' value '{os.environ[env_var]}' to '{format[k].__name__}'"
                    )

        # set last config value found
        try:
            ret[k] = next(c[k] for c in config[::-1] if k in c)
        except StopIteration:  # no matching key in config
            pass

        if ret[k] is not None:
            try:
                ret[k] = format[k](ret[k])
            except:
                raise ValueError(f"config '{kpath}' value '{ret[k]}' can't be cast to '{format[k].__name__}'")

    return ret


def load_file(config: Path):
    if not config.exists():
        raise FileExistsError(f"config file '{config}' doesn't exists.")

    if config.suffix == ".json":
        with open(config) as f:
            ret = json.load(f)
    else:
        raise RuntimeError("Unsupported config file. only [.py, .json] file types are supported.")

    return ret


def load_config(config: Union[CONFIG_TYPE, List[CONFIG_TYPE]] = None, environ=True):
    loaded = False
    if config is None or config == DEFAULT_CONFIG:
        # default config used as base configuration for _load_config
        config = []
    elif isinstance(config, dict):
        # flag indicating config was already loaded (and hence shuold be used as is)
        loaded = config.get("_loaded", False)
        config = config if loaded else [config]
    elif isinstance(config, list):
        for i, c in enumerate(config):
            if not isinstance(c, get_args(CONFIG_TYPE)):
                raise RuntimeError(f"onvalid config[{i}] element type. supported types: {_get_args_names(CONFIG_TYPE)}")
    elif isinstance(config, get_args(CONFIG_TYPE)):
        # list already handled as pass
        config = [config]
    else:
        raise RuntimeError(f"invalid config type. supported types: {_get_args_names(CONFIG_TYPE)}")

    if loaded:
        return config

    # convert config to dicts
    for i, c in enumerate(config):
        if isinstance(c, str):
            if c in CONFIG_SETS:
                # presets
                config[i] = CONFIG_SETS[c]
            else:
                config[i] = load_file(Path(c))
        elif isinstance(c, Path):
            config[i] = load_file(c)
        elif isinstance(c, dict):
            continue

    ret = _load_config(config, CONFIG_SETS[DEFAULT_CONFIG], environ=environ)
    ret["_loaded"] = True

    return ret


def get_config_set(config=DEFAULT_CONFIG):
    return copy.deepcopy(CONFIG_SETS[config])



from pytest import fixture
import os
import pytest
import copy

from .config import load_config
from .env import ATASKQ_CONFIG

if os.getenv("_PYTEST_RAISE", "0") != "0":

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call):
        raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo):
        raise excinfo.value


def truncate_query(table_name):
    return f"""DO $$
BEGIN
   IF EXISTS (SELECT 1 FROM information_schema.tables WHERE table_schema = 'public' AND table_name = '{table_name}') THEN
      TRUNCATE TABLE {table_name} CASCADE;
   END IF;
END $$;
"""


def drop_pg_tables(conn):
    assert "pg" in conn
    import psycopg2
    from ataskq.handler.postgresql import PostgresqlDBHandler

    connection = PostgresqlDBHandler.from_connection_str(conn)
    db_conn = psycopg2.connect(
        host=connection.host, database=connection.database, user=connection.user, password=connection.password
    )
    c = db_conn.cursor()
    c.execute(truncate_query("tasks"))
    c.execute(truncate_query("jobs"))
    db_conn.commit()
    db_conn.close()


@fixture
def config(tmp_path):
    config = load_config(ATASKQ_CONFIG or "test")
    conn = config["connection"]
    if "sqlite" in conn:
        config["connection"] = conn.format(tmp_path=tmp_path)
    elif "pg" in conn:
        # connect and clear all db tables
        drop_pg_tables(conn)
    elif "http" in conn:
        server_conn = os.getenv("TEST_SERVER_CONNECTION", "pg://postgres:postgres@localhost/postgres")
        assert "pg" in server_conn, "rest apigi test must be with posgres server"
        drop_pg_tables(server_conn)
    else:
        raise Exception(f"Unkown connection format '{conn}'")

    return copy.deepcopy(config)



from datetime import datetime, timedelta
from typing import List
from abc import abstractmethod
from datetime import datetime

from .handler import Handler, get_query_kwargs
from ..imodel import IModel
from .. import __schema_version__

CONNECTION_LOG = [None]


def set_connection_log(log):
    CONNECTION_LOG[0] = log


def transaction_decorator(exclusive=False):
    def decorator(func):
        def wrapper(self, *args, **kwargs):
            with self.connect() as conn:
                if CONNECTION_LOG[0] is not None:
                    conn.set_trace_callback(CONNECTION_LOG[0])

                c = conn.cursor()
                try:
                    if self.pragma_foreign_keys_on:
                        c.execute(self.pragma_foreign_keys_on)

                    if exclusive:
                        c.execute(self.begin_exclusive)
                    else:
                        c.execute("BEGIN")
                    # enable foreign keys for each connection (sqlite default is off)
                    # https://www.sqlite.org/foreignkeys.html
                    # Foreign key constraints are disabled by default (for backwards
                    # compatibility), so must be enabled separately for each database
                    # connection
                    ret = func(self, c, *args, **kwargs)
                    conn.commit()
                except Exception as e:
                    self.error(f"Failed to execute transaction '{type(e)}:{e}'. Rolling back")
                    conn.rollback()
                    raise e

            return ret

        return wrapper

    return decorator


def _field_with_order(f):
    if isinstance(f, (tuple, list)):
        if len(f) == 1:
            return f"{f[0]} ASC"
        assert len(f) == 2, "order_by tuple must be of format (field, order) where order default is ASC"

    else:
        return f"{f} ASC"
    return f"{f[0]} {f[1]}"


def order_query(order_by):
    if isinstance(order_by, str):
        return order_by

    assert isinstance(order_by, (tuple, list)), "order_by must be of type str, tuple or list"

    order_by = [_field_with_order(f) for f in order_by]
    order_by = ", ".join(order_by)

    return order_by


def expand_query_str(query_str, _where=None, _group_by=None, _order_by=None, _limit=None, _offset=None):
    if _where is not None:
        query_str += f" WHERE {_where}"

    if _group_by is not None:
        if not isinstance(_group_by, (list, tuple)):
            _group_by = [_group_by]
        query_str += f" GROUP BY {', '.join(_group_by)}"

    if _order_by is not None:
        query_str += f" ORDER BY {order_query(_order_by)}"

    if _limit is not None:
        query_str += f" LIMIT {_limit}"

    if _offset is not None:
        query_str += f" OFFSET {_offset}"

    return query_str


class DBHandler(Handler):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        if self.config["handler"]["db_init"]:
            self.init_db()

    @property
    def db_path(self):
        raise Exception(f"'{self.__class__.__name__}' db doesn't support db path property'")

    def count_all(self, model_cls: IModel, **kwargs):
        query_kwargs = get_query_kwargs(kwargs)
        count = self.count_query(model_cls, **query_kwargs)

        return count

    def get_all(self, model_cls: IModel, **kwargs) -> List[dict]:
        query_kwargs = get_query_kwargs(kwargs)
        rows, col_names, _ = self.select_query(model_cls, **query_kwargs)
        ret = [dict(zip(col_names, row)) for row in rows]

        return ret

    def get(self, model_cls: IModel, model_id) -> dict:
        rows, col_names, query_str = self.select_query(model_cls, _where=f"{model_cls.id_key()} = {model_id}")
        assert len(rows) != 0, f"no match found for '{model_cls.__name__}', query: '{query_str}'."
        assert len(rows) == 1, f"more than 1 row found for '{model_cls.__name__}', query: '{query_str}'."
        ret = [dict(zip(col_names, row)) for row in rows][0]

        return ret

    @property
    def pragma_foreign_keys_on(self):
        return None

    @property
    @abstractmethod
    def format_symbol(self):
        pass

    @property
    @abstractmethod
    def connection(self):
        pass

    @property
    @abstractmethod
    def bytes_type(self):
        pass

    @property
    @abstractmethod
    def primary_key(self):
        pass

    @property
    @abstractmethod
    def timestamp_type(self):
        pass

    @abstractmethod
    def timestamp(self, ts):
        pass

    @property
    @abstractmethod
    def begin_exclusive(self):
        pass

    @property
    @abstractmethod
    def for_update(self):
        pass

    @abstractmethod
    def connect(self):
        pass

    def _create(self, model_cls: IModel, **ikwargs) -> int:
        model_id = self.create_bulk(model_cls, [ikwargs])[0]

        return model_id

    @transaction_decorator()
    def _create_bulk(self, c, model_cls: IModel, ikwargs: List[dict]) -> List[int]:
        # todo: consolidate all ikwargs with same keys to single insert command
        model_ids = []
        for v in ikwargs:
            d = {k: v for k, v in v.items() if model_cls.id_key() not in k}
            keys = list(d.keys())
            values = list(d.values())
            c.execute(
                f'INSERT INTO {model_cls.table_key()} ({", ".join(keys)}) VALUES ({", ".join([self.format_symbol] * len(keys))}) RETURNING {model_cls.id_key()}',
                values,
            )
            model_id = c.fetchone()[0]
            model_ids.append(model_id)

        return model_ids

    @transaction_decorator()
    def _update(self, c, model_cls: IModel, model_id, **ikwargs):
        if len(ikwargs) == 0:
            return

        insert = ", ".join([f"{k} = {self.format_symbol}" for k in ikwargs.keys()])
        values = list(ikwargs.values())
        c.execute(f"UPDATE {model_cls.table_key()} SET {insert} WHERE {model_cls.id_key()} = {model_id};", values)

    @transaction_decorator()
    def update_all(self, c, model_cls: IModel, where: str = None, **ikwargs):
        if len(ikwargs) == 0:
            return

        insert = ", ".join([f"{k} = {self.format_symbol}" for k in ikwargs.keys()])
        values = list(ikwargs.values())

        query_str = f"UPDATE {model_cls.table_key()} SET {insert}"

        if where:
            query_str += f" WHERE {where}"

        c.execute(query_str, values)

    @abstractmethod
    def delete(self, model_cls: IModel, model_id: int):
        pass

    @transaction_decorator()
    def delete_all(self, c, model_cls: IModel, **kwargs):
        query_kwargs = get_query_kwargs(kwargs)
        query_str = f"DELETE FROM {model_cls.table_key()}"

        if "_where" in query_kwargs:
            query_str += f" WHERE {query_kwargs['_where']}"

        c.execute(query_str)

    @transaction_decorator()
    def delete(self, c, model_cls: IModel, model_id: int):
        c.execute(f"DELETE FROM {model_cls.table_key()} WHERE {model_cls.id_key()} = {model_id}")

    @transaction_decorator(exclusive=True)
    def init_db(self, c):
        from ..models import EStatus

        # Create schema version table if not exists
        c.execute("CREATE TABLE IF NOT EXISTS schema_version (" "version INTEGER PRIMARY KEY" ")")
        c.execute("SELECT * FROM schema_version")
        current_schema_version = c.fetchone()
        if current_schema_version is None:
            c.execute(f"INSERT INTO schema_version (version) VALUES ({__schema_version__})")
        else:
            current_schema_version = current_schema_version[0]
            assert (
                current_schema_version == __schema_version__
            ), f"Schema version mismatch, current schema version is {current_schema_version} while code schema version is {__schema_version__}"

        # Create jobs table if not exists
        c.execute(
            "CREATE TABLE IF NOT EXISTS jobs ("
            f"job_id {self.primary_key}, "
            "name TEXT, "
            "description TEXT, "
            "priority REAL DEFAULT 0"
            #   "summary_cookie_keys JSON"
            ")"
        )

        # Create tasks table if not exists
        statuses = ", ".join([f'"{a}"' for a in EStatus])
        c.execute(
            f"CREATE TABLE IF NOT EXISTS tasks ("
            f"task_id {self.primary_key}, "
            "name TEXT, "
            "level REAL, "
            "entrypoint TEXT NOT NULL, "
            f"targs {self.bytes_type}, "
            f"status TEXT ,"  # CHECK(status in ({statuses})),
            f"take_time {self.timestamp_type}, "
            f"start_time {self.timestamp_type}, "
            f"done_time {self.timestamp_type}, "
            f"pulse_time {self.timestamp_type}, "
            "description TEXT, "
            #   "summary_cookie JSON, "
            "job_id INTEGER NOT NULL, "
            "CONSTRAINT fk_job_id FOREIGN KEY (job_id) REFERENCES jobs(job_id) ON DELETE CASCADE"
            ")"
        )

    @transaction_decorator()
    def count_query(self, c, model_cls: IModel, _where: str = None, _limit: int = None, _offset: int = 0):
        if _limit is None:
            _limit = self.config["api"]["limit"]
        query_str = f"SELECT COUNT(*) FROM {model_cls.table_key()}"
        query_str = expand_query_str(query_str, _where=_where, _limit=_limit, _offset=_offset)

        c.execute(query_str)

        row = c.fetchone()
        return row[0]

    @transaction_decorator()
    def select_query(
        self,
        c,
        model_cls: IModel,
        _where: str = None,
        _order_by=None,
        _limit: int = None,
        _offset: int = 0,
    ):
        if _limit is None:
            _limit = self.config["api"]["limit"]
        query_str = f"SELECT * FROM {model_cls.table_key()}"
        if _order_by is None:
            _order_by = f"{model_cls.table_key()}.{model_cls.id_key()} ASC"
        query_str = expand_query_str(query_str, _where=_where, _order_by=_order_by, _limit=_limit, _offset=_offset)

        c.execute(query_str)
        rows = c.fetchall()
        col_names = [description[0] for description in c.description]

        return rows, col_names, query_str

    ##################
    # Custom Queries #
    ##################

    @transaction_decorator(exclusive=True)
    def take_next_task(self, c, job_id: int = None, level_start: int = None, level_stop: int = None):
        # imported here to avoid circular dependency
        from ..models import Task, EStatus
        from .handler import EAction

        # todo: add FOR UPDATE in the queries postgresql
        level_query = ""
        if level_start:
            level_query += f" AND level >= {level_start}"
        if level_stop:
            level_query += f" AND level < {level_stop}"

        # get pending task with minimum level
        job_query = ""
        if job_id is not None:
            job_query += f" AND job_id = {job_id}"

        query = (
            f"SELECT * FROM tasks WHERE status IN ('{EStatus.PENDING}'){job_query}{level_query} AND level = "
            f"(SELECT MIN(level) FROM tasks WHERE status IN ('{EStatus.PENDING}'){job_query}{level_query})"
            f" ORDER BY job_id ASC, task_id ASC {self.for_update}"
        )
        query = query.strip()

        c.execute(query)
        row = c.fetchone()
        if row is None:
            ptask = None
        else:
            col_names = [description[0] for description in c.description]
            ptask = self.from_interface(Task, dict(zip(col_names, row)))

        # get running task with minimum level
        query = (
            f"SELECT * FROM tasks WHERE status IN ('{EStatus.RUNNING}'){job_query}{level_query} AND level = "
            f"(SELECT MIN(level) FROM tasks WHERE status IN ('{EStatus.RUNNING}'){job_query}{level_query})"
            f" {self.for_update}"
        )
        query = query.strip()
        c.execute(query)
        row = c.fetchone()
        if row is None:
            rtask = None
        else:
            col_names = [description[0] for description in c.description]
            rtask = self.from_interface(Task, dict(zip(col_names, row)))

        action = None
        if ptask is None and rtask is None:
            # no more pending task, no more running tasks
            action = EAction.STOP
        elif ptask is None and rtask is not None:
            # no more pending tasks, tasks still running
            action = EAction.WAIT
        elif ptask is not None and rtask is None:
            # pending task next, no more running tasks
            action = EAction.RUN_TASK
        elif ptask is not None and rtask is not None:
            if ptask.level > rtask.level:
                # pending task with level higher than running (wait for running to end)
                action = EAction.WAIT
            elif rtask.level > ptask.level:
                # should never happend
                # running task with level higher than pending task (warn and take next task)
                self.warning(
                    f"Running task with level higher than pending detected, taking pending. running id: {rtask.task_id}, pending id: {ptask.task_id}."
                )
                action = EAction.RUN_TASK
            else:
                action = EAction.RUN_TASK

        if action == EAction.RUN_TASK:
            now = datetime.now()
            c.execute(
                f"UPDATE tasks SET status = '{EStatus.RUNNING}', take_time = {self.timestamp(now)}, pulse_time = {self.timestamp(now)} WHERE task_id = {ptask.task_id};"
            )
            ptask.status = EStatus.RUNNING
            ptask.take_time = now
            ptask.pulse_time = now
            task = ptask
        elif action == EAction.WAIT:
            task = None
        elif action == EAction.STOP:
            task = None
        else:
            raise RuntimeError(f"Unsupported action '{EAction}'")

        return action, task

    @transaction_decorator()
    def tasks_status(
        self,
        c,
        **kwargs,
    ):
        from ..models import EStatus

        # todo add group by to get_query_kwargs

        if kwargs.get("_limit") is None:
            kwargs["_limit"] = self.config["api"]["limit"]
        if kwargs.get("_order_by") is None:
            kwargs["_order_by"] = "name ASC"
        if kwargs.get("_group_by") is None:
            kwargs["_group_by"] = ("level", "name")
        query_kwargs = get_query_kwargs(kwargs)

        query_str = (
            "SELECT level, name,"
            "COUNT(*) as total, "
            + ",".join([f"SUM(CASE WHEN status = '{status}' THEN 1 ELSE 0 END) AS {status} " for status in EStatus])
            + "FROM tasks"
        )
        query_str = expand_query_str(query_str, **query_kwargs)

        c.execute(query_str)
        rows = c.fetchall()
        col_names = [description[0] for description in c.description]

        ret = [dict(zip(col_names, row)) for row in rows]
        return ret

    @transaction_decorator()
    def jobs_status(self, c, _order_by: str = None, _limit: int = None, _offset: int = 0):
        from ..models import EStatus

        if _limit is None:
            _limit = self.config["api"]["limit"]

        query_str = (
            "SELECT jobs.job_id, jobs.name, jobs.description, jobs.priority, "
            "COUNT(*) as tasks, "
            + ", ".join([f"SUM(CASE WHEN status = '{status}' THEN 1 ELSE 0 END) AS {status}" for status in EStatus])
            + f" FROM jobs "
            "LEFT JOIN tasks ON jobs.job_id = tasks.job_id "
            "GROUP BY jobs.job_id"
        )

        if _order_by is None:
            _order_by = "jobs.job_id DESC"

        query_str = expand_query_str(query_str, _order_by=_order_by, _limit=_limit, _offset=_offset)

        c.execute(query_str)
        rows = c.fetchall()
        col_names = [description[0] for description in c.description]

        ret = [dict(zip(col_names, row)) for row in rows]
        return ret

    @transaction_decorator()
    def fail_pulse_timeout_tasks(self, c, timeout_sec=None):
        from ..models import EStatus

        if timeout_sec is None:
            return

        # set timeout tasks
        last_valid_pulse = datetime.now() - timedelta(seconds=timeout_sec)
        c.execute(
            f"UPDATE tasks SET status = '{EStatus.FAILURE}' WHERE pulse_time < {self.timestamp(last_valid_pulse)} AND status NOT IN ('{EStatus.SUCCESS}', '{EStatus.FAILURE}');"
        )



from multiprocessing import Lock
import os
import time
from datetime import datetime


def write_to_file(filepath, text, sleep=None):
    if sleep is not None:
        time.sleep(sleep)
    text = text.replace(r"@{pid}", f"{os.getpid()}").replace(r"@{now}", f"{datetime.now()}")
    with open(filepath, "a") as f:
        f.write(text)


__lock__ = Lock()


def write_to_file_mp_lock(filepath, text, sleep=None):
    if sleep is not None:
        time.sleep(sleep)
    text = text.replace(r"@{pid}", f"{os.getpid()}").replace(r"@{now}", f"{datetime.now()}")
    with __lock__:
        with open(filepath, "a") as f:
            f.write(text)



import multiprocessing
from typing import Union
import pickle
import logging
from importlib import import_module
from multiprocessing import Process
import time
from typing import List
from datetime import datetime


from .logger import Logger
from .models import EStatus, Job, Task
from .monitor import MonitorThread
from .handler import Handler, DBHandler, from_config, EAction
from .config import load_config


def targs(*args, **kwargs):
    return (args, kwargs)


class TaskQ(Logger):
    def __init__(
        self,
        job_id=None,
        handler: Handler = None,
        logger: Union[str, logging.Logger, None] = None,
        config=None,
    ) -> None:
        super().__init__(logger)

        # init config
        self._config = load_config(config)

        # init db handler
        # todo: handler shouldn't get job_id
        if handler is None:
            self._handler: Handler = from_config(
                config=config,
                logger=self._logger,
            )
        else:
            self._handler = handler

        # get job
        if job_id is not None:
            try:
                job = Job.get(job_id, _handler=self._handler)
            except Exception as ex:
                raise Exception(f"Failed to fetch job (job_id={job_id}).") from ex
        else:
            job = None
        self._job = job

        self._running = False

    @property
    def config(self):
        return self._config

    @property
    def job(self):
        return self._job

    @property
    def job_id(self):
        return self._job.job_id

    @property
    def handler(self):
        return self._handler

    @property
    def task_wait_interval(self):
        return self._task_pull_interval

    @property
    def monitor_pulse_interval(self):
        return self._monitor_pulse_interval

    def clear_job(self):
        self._job = None

    def set_job(self, job_id):
        job = Job.get(job_id, self._handler)
        self._handler.set_job_id(job.job_id)  # todo remove - handler should need job id ...
        self._job = job

    def create_job(self, name=None, description=None):
        assert self._job is None, "job already assigned to current taskq, run clear_job to create new job."

        job = Job(name=name, description=description).create(_handler=self._handler)
        self._job = job

        if self.config["db"]["max_jobs"] is not None:
            # keep max jbos
            Job.delete_all(
                _where=f"job_id NOT IN (SELECT job_id FROM jobs ORDER BY job_id DESC limit {self.config['db']['max_jobs']})",
                _handler=self.handler,
            )

        return self

    def delete_job(self):
        self.job.delete(_handler=self.handler)
        self._job = None

    def get_tasks(self) -> List[Task]:
        return self.job.get_tasks(self._handler)

    def add_tasks(self, tasks):
        self.job.add_tasks(tasks, _handler=self._handler)

        return self

    def update_task_start_time(self, task: Task, start_time: datetime = None):
        if start_time is None:
            start_time = datetime.now()

        task.update(start_time=start_time, _handler=self._handler)

    def update_task_status(self, task: Task, status: EStatus, timestamp: datetime = None):
        if timestamp is None:
            timestamp = datetime.now()

        if status == EStatus.RUNNING:
            # for running task update pulse_time
            task.update(_handler=self._handler, status=status, pulse_time=timestamp)
        elif status == EStatus.SUCCESS or status == EStatus.FAILURE:
            # for done task update pulse_time and done_time time as well
            task.update(_handler=self._handler, status=status, pulse_time=timestamp, done_time=timestamp)
        else:
            raise RuntimeError(f"Unsupported status '{status}' for status update")

    def count_pending_tasks_below_level(self, level):
        ret = Task.count_all(
            _where=f"job_id = {self.job_id} AND level < {level} AND status in ('{EStatus.PENDING}')",
            _handler=self._handler,
        )
        return ret

    def _run_task(self, task: Task):
        self.info(f"Running task '{task}'")

        # get entry point func to execute
        ep = task.entrypoint
        if ep == "ataskq.skip_run_task":
            self.info(f"task '{task}' is marked as 'skip_run_task', skipping run task.")
            return

        assert "." in ep, "entry point must be inside a module."
        module_name, func_name = ep.rsplit(".", 1)
        try:
            m = import_module(module_name)
        except ImportError as ex:
            raise RuntimeError(f"Failed to load module '{module_name}'. Exception: '{ex}'")
        assert hasattr(
            m, func_name
        ), f"failed to load entry point, module '{module_name}' doen't have func named '{func_name}'."
        func = getattr(m, func_name)
        assert callable(func), f"entry point is not callable, '{module_name},{func}'."

        # get targs
        if task.targs is not None:
            try:
                targs = pickle.loads(task.targs)
            except Exception as ex:
                if self.config["run"]["raise_exception"]:  # for debug purposes only
                    self.warning("Getting tasks args failed.")
                    self.update_task_status(task, EStatus.FAILURE)
                    raise ex

                self.warning("Getting tasks args failed.", exc_info=True)
                self.update_task_status(task, EStatus.FAILURE)

                return
        else:
            targs = ((), {})

        # update task start time
        self.update_task_start_time(task)

        # run task
        monitor = MonitorThread(task, self, pulse_interval=self.config["monitor"]["pulse_interval"])
        monitor.start()

        try:
            func(*targs[0], **targs[1])
            status = EStatus.SUCCESS
        except Exception as ex:
            msg = f"Running task '{task}' failed with exception."
            if self.config["run"]["raise_exception"]:  # for debug purposes only
                self.warning(msg)
                self.update_task_status(task, EStatus.FAILURE)
                monitor.stop()
                monitor.join()
                raise ex

            self.warning(msg, exc_info=True)
            status = EStatus.FAILURE

        monitor.stop()
        monitor.join()
        self.update_task_status(task, status)

    def _take_next_task(self, level=None):
        level_start = level.start if level is not None else None
        level_stop = level.stop if level is not None else None

        job_id = self.job_id if self.job is not None else None
        return self._handler.take_next_task(job_id=job_id, level_start=level_start, level_stop=level_stop)

    def _run(self, level):
        self.info(f"Started task pulling loop.")

        # check for error code
        task_pull_start = time.time()
        while True:
            # if the taskq handler is db handler, the taskq performs background tasks before each run
            if self.config["run"]["fail_pulse_timeout"] and isinstance(self._handler, DBHandler):
                self._handler.fail_pulse_timeout_tasks(self.config["monitor"]["pulse_timeout"])
            # grab tasks and set them in Q
            action, task = self._take_next_task(level)

            # handle no task available
            if not self.config["run"]["run_forever"] and action == EAction.STOP:
                break
            if action == EAction.RUN_TASK:
                self._run_task(task)
            elif action == EAction.WAIT or action == EAction.STOP:
                if (
                    wait_timeout := self.config["run"]["wait_timeout"]
                ) is not None and time.time() - task_pull_start > wait_timeout:
                    raise Exception(f"task pull timeout of '{wait_timeout}' sec reached.")

                self.info(f'Task pulling loop - waiting for {self.config["run"]["pull_interval"]} sec')
                time.sleep(self.config["run"]["pull_interval"])
            else:
                raise Exception(f"Unsupported action {action}")

    def assert_level(self, level):
        if isinstance(level, int):
            level = range(level, level + 1)
        elif isinstance(level, (list, tuple)):
            assert len(level) == 2, "level of type list or tuple must have length of 2"
            level = range(level[0], level[1])
        else:
            assert isinstance(level, range), "level must be int, list, tuple or range"

        # check all task < level.start are done
        count = self.count_pending_tasks_below_level(level.start)
        assert count == 0, f"all tasks below level must be done before running tasks at levels {level}"

        return level

    def run(self, concurrency=None, level=None):
        self.info(f"Start running with connection {self.config['connection']}")
        if level is not None:
            level = self.assert_level(level)

        self._running = True

        # default to run in current process
        if concurrency is None:
            self._run(level)
            self._running = False
            return

        assert isinstance(concurrency, (int, float))

        if isinstance(concurrency, float):
            assert 0.0 <= concurrency <= 1.0
            nprocesses = int(multiprocessing.cpu_count() * concurrency)
        elif concurrency < 0:
            nprocesses = multiprocessing.cpu_count() - concurrency
        else:
            nprocesses = concurrency

        # set processes and Q
        processes = [Process(target=self._run, args=(level,)) for i in range(nprocesses)]
        [p.start() for p in processes]

        # join all processes
        [p.join() for p in processes]

        # log failed processes
        fail = False
        for p in processes:
            if p.exitcode != 0:
                fail = True
                self.error(f"Process '{p.pid}' failed with exitcode '{p.exitcode}'")

        self._running = False

        if self.config["run"]["raise_exception"] and fail:
            raise Exception("Some processes failed, see logs for details")

        return self



from typing import Union
import logging


class Logger:
    def __init__(self, logger: Union[str, logging.Logger, None]):
        if logger is None:
            self._logger = logging.getLogger("ataskq")
        elif isinstance(logger, str):
            self._logger = logging.getLogger(logger)
        else:
            self._logger = logger

    def exception(self, *args, **kwargs):
        self._logger.exception(*args, **kwargs)

    def critical(self, *args, **kwargs):
        self._logger.critical(*args, **kwargs)

    def error(self, *args, **kwargs):
        self._logger.error(*args, **kwargs)

    def warning(self, *args, **kwargs):
        self._logger.warning(*args, **kwargs)

    def info(self, *args, **kwargs):
        self._logger.info(*args, **kwargs)

    def debug(self, *args, **kwargs):
        self._logger.debug(*args, **kwargs)



from threading import Thread, Event

from .models import Task, EStatus


class MonitorThread(Thread):
    # task runner is .task_runner TaskRunner, avoiding circular import
    def __init__(self, task: Task, ataskq, pulse_interval: float = 60) -> None:
        from .taskq import TaskQ  # here to avoid circular dependency

        super().__init__(daemon=True)
        self._stop_event = Event()
        self._task = task
        self._ataskq: TaskQ = ataskq
        self._pulse_interval = pulse_interval

    def run(self) -> None:
        self._ataskq.info(f"Running monitor thread for task '{self._task}'")
        while not self._stop_event.is_set():
            self._ataskq.update_task_status(self._task, EStatus.RUNNING)
            self._stop_event.wait(self._pulse_interval)

    def stop(self):
        self._stop_event.set()



class CounterKWArg:
    def __init__(self, val=0, name="counter") -> None:
        self._val = val
        self._name = name

    @property
    def val(self):
        return self._val

    @property
    def name(self):
        return self._name

    def count(self):
        ret = self._val
        self._val += 1

        return ret


def counter_kwarg():
    return CounterKWArg()


def counter_task(counter: CounterKWArg = None, print_counter=False):
    assert counter is not None, "counter value must be provided"

    if print_counter:
        print(f"counter name: {counter.name}, val: {counter.val}")
    counter.count()



from abc import abstractmethod
from typing import Union, List, Dict
from datetime import datetime
from enum import Enum
import copy

from ..env import ATASKQ_CONFIG
from ..logger import Logger
from ..config import load_config
from ..imodel import IModel, IModelSerializer

__STRTIME_FORMAT__ = "%Y-%m-%d %H:%M:%S.%f"


def get_query_kwargs(kwargs):
    # todo: the k=v for the kwargs should be interface dependent similar to insert
    ret = {}
    _where = ""
    if "_where" in kwargs:
        _where += kwargs["_where"]
    for k, v in kwargs.items():
        if k == "_where":
            continue
        if k in ["_group_by", "_order_by", "_limit", "_offset"]:
            ret[k] = v
            continue
        if v is None:
            continue
        _where += f"{_where and ' AND '}{k}={v}"

    _where = _where or None
    if _where:
        ret["_where"] = _where

    return ret


def to_datetime(string: Union[str, datetime, None]):
    if string is None:
        return None
    elif isinstance(string, datetime):
        return string

    return datetime.strptime(string, __STRTIME_FORMAT__)


def from_datetime(time: datetime):
    return time.strftime(__STRTIME_FORMAT__)


class EAction(str, Enum):
    RUN_TASK = "run_task"
    WAIT = "wait"
    STOP = "stop"

    def __str__(self) -> str:
        return self.value


class Handler(IModelSerializer, Logger):
    def __init__(self, config=ATASKQ_CONFIG, logger: Logger = None):
        Logger.__init__(self, logger)

        # init config
        self._config = load_config(config)
        self._connection = self.from_connection_str(self._config["connection"])

    @property
    def config(self):
        return self._config

    @staticmethod
    @abstractmethod
    def from_connection_str(conn):
        pass

    ######################
    # interface handlers #
    ######################

    @classmethod
    def i2m(cls, model_cls, kwargs: Union[dict, List[dict]]) -> Union[dict, List[dict]]:
        """interface to model"""
        return model_cls.i2m(kwargs, cls)

    @classmethod
    def from_interface(cls, model_cls: IModel, kwargs: Union[dict, List[dict]]) -> Union[IModel, List[IModel]]:
        return model_cls.from_interface(kwargs, cls)

    @classmethod
    def m2i(cls, model_cls: IModel, kwargs: Union[dict, List[dict]]) -> Union[dict, List[dict]]:
        """modle to interface"""
        return model_cls.m2i(kwargs, cls)

    @classmethod
    def to_interface(cls, model: IModel) -> IModel:
        return model.to_interface(cls)

    ########
    # CRUD #
    ########

    @abstractmethod
    def _create(self, model_cls: IModel, **ikwargs: dict):
        pass

    @abstractmethod
    def _create_bulk(self, model_cls: IModel, ikwargs: List[dict]):
        pass

    @abstractmethod
    def delete_all(self, model_cls: IModel, **kwargs):
        pass

    @abstractmethod
    def delete(self, model_cls: IModel, model_id: int):
        pass

    @abstractmethod
    def count_all(self, model_cls: IModel, **kwargs) -> int:
        pass

    @abstractmethod
    def get_all(self, model_cls: IModel, **kwargs) -> List[dict]:
        pass

    @abstractmethod
    def get(self, model_cls: IModel, model_id: int) -> dict:
        pass

    def create(self, model_cls: IModel, **mkwargs) -> int:
        assert (
            model_cls.id_key() not in mkwargs
        ), f"id '{model_cls.id_key()}' can't be passed to create '{model_cls.__name__}({model_cls.table_key()})'"
        ikwargs = self.m2i(model_cls, mkwargs)
        model_id = self._create(model_cls, **ikwargs)

        return model_id

    def create_bulk(self, model_cls: IModel, mkwargs: List[dict]) -> List[int]:
        for i, v in enumerate(mkwargs):
            assert (
                model_cls.id_key() not in v
            ), f"item [{i}]: id '{model_cls.id_key()}' can't be passed to create '{model_cls.__name__}({model_cls.table_key()})'"
        ikwargs = self.m2i(model_cls, mkwargs)
        model_ids = self._create_bulk(model_cls, ikwargs)

        return model_ids

    @abstractmethod
    def _update(self, model_cls: IModel, model_id: int, **ikwargs):
        pass

    @abstractmethod
    def update_all(self, model_cls: IModel, where: str = None, **ikwargs):
        pass

    def update(self, model_cls: IModel, model_id: int, **mkwargs):
        assert model_id is not None, f"{model_cls} must have assigned '{model_cls.id_key()}' for update"
        ikwargs = self.m2i(model_cls, mkwargs)
        self._update(model_cls, model_id, **ikwargs)

    ##########
    # Custom #
    ##########
    @abstractmethod
    def take_next_task(self, job_id=None, level_start: int = None, level_stop: int = None) -> tuple:
        pass

    @abstractmethod
    def tasks_status(self, job_id=None, _order_by: str = None, _limit: int = None, _offset: int = 0) -> List[dict]:
        pass

    @abstractmethod
    def jobs_status(self, _order_by: str = None, _limit: int = None, _offset: int = 0) -> List[dict]:
        pass


__HANDLERS__: Dict[str, object] = dict()


def register_handler(name, handler: Handler):
    """register interface handlers"""
    __HANDLERS__[name] = handler


def unregister_handler(name):
    """register interface handlers"""
    return __HANDLERS__.pop(name)


def get_handler(name=None, assert_registered=False):
    """get registered interface handlers"""

    if len(__HANDLERS__) == 0:
        if assert_registered:
            raise RuntimeError("No registered interface handlers")
        return None
    elif len(__HANDLERS__) == 1:
        return list(__HANDLERS__.values())[0]
    else:
        assert (
            name is not None
        ), f"more than 1 type hander registered, please specify handler name. registered handlers: {list(__HANDLERS__.keys())}"
        assert (
            name in __HANDLERS__
        ), f"no handler named '{name}' is registered. registered handlers: {list(__HANDLERS__.keys())}"
        return __HANDLERS__[name]


def from_config(config=ATASKQ_CONFIG, **kwargs) -> Handler:
    # expand config in factory and not inside handler
    config = load_config(config)
    conn = config["connection"]

    sep = "://"
    sep_index = conn.find(sep)
    if sep_index == -1:
        raise RuntimeError("connection must be of format <type>://<connection string>")
    handler_type = conn[:sep_index]

    # validate connectino
    if not handler_type:
        raise RuntimeError("missing handler type, connection must be of format <type>://<connection string>")

    connection_str = conn[sep_index + len(sep) :]
    if not connection_str:
        raise RuntimeError("missing connection string, connection must be of format <type>://<connection string>")

    # get db type handler
    kwargs["config"] = config
    if handler_type == "sqlite":
        from .sqlite3 import SQLite3DBHandler

        handler = SQLite3DBHandler(**kwargs)
    elif handler_type == "pg":
        from .postgresql import PostgresqlDBHandler

        handler = PostgresqlDBHandler(**kwargs)
    elif handler_type == "http" or handler_type == "https":
        from .rest_handler import RESTHandler

        handler = RESTHandler(**kwargs)
    else:
        raise Exception(f"unsupported handler type '{handler_type}', type must be one of ['sqlite', 'pg', 'http']")

    return handler



DEFAULT_CONFIG = "standalone"

CONFIG_FORMAT = {
    "connection": str,  # ATASKQ_CONNECTION
    "run": {
        "wait_timeout": float,
        "pull_interval": float,
        "fail_pulse_timeout": bool,
        "raise_exception": bool,
        "run_forever": bool,
    },
    "handler": {
        "db_init": bool,
    },
    "db": {
        "max_jobs": int,
    },
    "monitor": {
        "pulse_interval": float,
        "pulse_timeout": float,
    },
    "background": {
        "pulse_timeout_interval": float,
    },
    "api": {
        "limit": int,
    },
}

CONFIG_SETS = {
    "standalone": {
        "connection": "sqlite://ataskq.db.sqlite3",
        "run": {
            "wait_timeout": None,
            "pull_interval": 15,
            "fail_pulse_timeout": True,
            "raise_exception": False,
            "run_forever": False,
        },
        "handler": {
            "db_init": True,
        },
        "db": {
            "max_jobs": None,
        },
        "monitor": {
            "pulse_interval": 15,
            "pulse_timeout": 60 * 5,
        },
        "background": {
            "pulse_timeout_interval": 60,
        },
        "api": {
            "limit": 100,
        },
    },
    "test": {
        "connection": "sqlite://{tmp_path}/ataskq.db.sqlite3",
        "run": {
            "pull_interval": 0.3,
        },
    },
    "client": {
        "connection": "http://localhost:8080",
        "run": {
            "fail_pulse_timeout": False,
        },
        "handler": {
            "db_init": False,
        },
    },
    "server": {
        "run": {
            "fail_pulse_timeout": False,
        },
        "handler": {
            "db_init": False,
        },
    },
}



from pathlib import Path
from datetime import datetime, timedelta
from copy import copy
from multiprocessing import Process
import time

import pytest

from . import TaskQ, Job, Task, targs, EStatus
from .handler import DBHandler
from .handler import EAction, from_config

from .tasks_utils import dummy_args_task, write_to_file


def non_decreasing(L):
    return all(x <= y for x, y in zip(L, L[1:]))


def non_increasing(L):
    return all(x >= y for x, y in zip(L, L[1:]))


def monotonic(L):
    return non_decreasing(L) or non_increasing(L)


@pytest.fixture
def jtaskq(config) -> TaskQ:
    return TaskQ(config=config).create_job()


def test_create_job(config):
    taskq = TaskQ(config=config).create_job()
    assert isinstance(taskq, TaskQ)

    conn = config["connection"]
    if "sqlite" in conn:
        assert Path(taskq.handler.db_path).exists()
        assert Path(taskq.handler.db_path).is_file()
    elif "pg" in conn:
        pass
    elif "http" in conn:
        pass
    else:
        raise Exception(f"unknown db type in connection string '{conn}'")


def test_job_default_name(config):
    job = TaskQ(config=config).create_job().job
    assert job.name is None


def test_job_custom_name(config):
    job = TaskQ(config=config).create_job(name="my_job").job
    assert job.name == "my_job"


def test_task_job_delete_cascade(config):
    # test that deleting a job deletes all its tasks
    handler = from_config(config)
    taskq1: TaskQ = TaskQ(handler=handler).create_job(name="job1")
    taskq2: TaskQ = TaskQ(handler=handler).create_job(name="job2")
    assert Job.count_all(_handler=handler) == 2

    taskq1.add_tasks(
        [
            Task(entrypoint=""),
            Task(entrypoint=""),
            Task(entrypoint=""),
        ]
    )
    assert Task.count_all(_handler=handler) == 3
    assert len(taskq1.get_tasks()) == 3

    taskq2.add_tasks(
        [
            Task(entrypoint=""),
            Task(entrypoint=""),
        ]
    )
    assert Task.count_all(_handler=handler) == 5
    assert len(taskq1.get_tasks()) == 3
    assert len(taskq2.get_tasks()) == 2

    taskq2.delete_job()
    assert len(Job.get_all(_handler=handler)) == 1

    assert Task.count_all(_handler=handler) == 3
    assert len(taskq1.get_tasks()) == 3

    taskq1.delete_job()
    assert Task.count_all(_handler=handler) == 0


def test_update_task_start_time(jtaskq):
    in_task1 = Task(entrypoint=dummy_args_task, level=1, name="task1")

    jtaskq.add_tasks(
        [
            in_task1,
        ]
    )

    tasks = jtaskq.get_tasks()
    assert len(tasks) == 1  # sanity
    task: Task = tasks[0]

    # update db with task time (also updates task inplace)
    start_time = datetime.now()
    start_time = start_time.replace(microsecond=0)  # test to seconds resolution
    jtaskq.update_task_start_time(task, start_time)
    assert task.start_time == start_time

    # get task form db and validate
    tasks = jtaskq.get_tasks()
    assert len(tasks) == 1  # sanity
    task: Task = tasks[0]
    assert task.start_time == start_time


def test_update_task_status(jtaskq):
    in_task1 = Task(entrypoint=dummy_args_task, level=1, name="task1")

    jtaskq.add_tasks(
        [
            in_task1,
        ]
    )

    tasks = jtaskq.get_tasks()
    assert len(tasks) == 1  # sanity
    task: Task = tasks[0]
    assert task.status == EStatus.PENDING

    now = datetime.now()
    # update db with task status (also updates task inplace)
    jtaskq.update_task_status(task, EStatus.RUNNING, timestamp=now)
    assert task.status == EStatus.RUNNING
    assert task.pulse_time == now
    assert task.done_time is None
    # get task form db and validate
    tasks = jtaskq.get_tasks()
    assert len(tasks) == 1  # sanity
    task: Task = tasks[0]
    assert task.status == EStatus.RUNNING
    assert task.pulse_time == now
    assert task.done_time is None

    # update db with task status (also updates task inplace)
    jtaskq.update_task_status(task, EStatus.SUCCESS, timestamp=now)
    assert task.status == EStatus.SUCCESS
    assert task.pulse_time == now
    assert task.done_time == now
    # get task form db and validate
    tasks = jtaskq.get_tasks()
    assert len(tasks) == 1  # sanity
    task: Task = tasks[0]
    assert task.status == EStatus.SUCCESS
    assert task.pulse_time == now
    assert task.done_time == now


def _compare_tasks(task1: Task, task2: Task, job_id=None):
    if job_id:
        assert task1.job_id == job_id
    assert task1.job_id == task2.job_id
    assert task1.name == task2.name
    assert task1.description == task2.description
    assert task1.level == task2.level
    assert task1.entrypoint == task2.entrypoint
    assert task1.targs == task2.targs


def test_get_tasks(jtaskq):
    in_task1 = Task(entrypoint=dummy_args_task, level=1, name="task1")
    in_task2 = Task(entrypoint=dummy_args_task, level=2, name="task2")
    in_task3 = Task(entrypoint=dummy_args_task, level=3, name="task3")

    jtaskq.add_tasks(
        [
            in_task3,
            in_task2,
            in_task1,
        ]
    )

    tasks = jtaskq.get_tasks()
    assert len(tasks) == 3
    for t in tasks:
        if t.level == 1:
            _compare_tasks(in_task1, t)
        elif t.level == 2:
            _compare_tasks(in_task2, t)
        elif t.level == 3:
            _compare_tasks(in_task3, t)


def test_take_next_task_sanity(jtaskq):
    in_task1 = Task(entrypoint=dummy_args_task, level=1, name="task1")

    jtaskq.add_tasks(
        [
            in_task1,
        ]
    )

    action, task = jtaskq._take_next_task(level=None)
    assert action == EAction.RUN_TASK
    _compare_tasks(in_task1, task)


def test_take_next_task(jtaskq):
    in_task1 = Task(entrypoint=dummy_args_task, level=1, name="task1")
    in_task2 = Task(entrypoint=dummy_args_task, level=2, name="task2")
    in_task3 = copy(in_task1)

    jtaskq.add_tasks(
        [
            in_task2,
            in_task1,
            in_task3,
        ]
    )

    tids = []

    action, task = jtaskq._take_next_task(level=None)
    assert action == EAction.RUN_TASK
    assert task.task_id not in tids
    tids.append(task.task_id)
    _compare_tasks(in_task1, task)

    action, task = jtaskq._take_next_task(level=None)
    assert action == EAction.RUN_TASK
    assert task.task_id not in tids
    tids.append(task.task_id)
    _compare_tasks(in_task3, task)  # note in_task3 is copy of in_task1

    action, task = jtaskq._take_next_task(level=None)
    assert action == EAction.WAIT
    assert task is None


def test_take_next_task_2_jobs(config):
    # todo: test should ne under ataskq
    handler = from_config(config)
    taskq1: TaskQ = TaskQ(handler=handler).create_job(name="job1")
    taskq2: TaskQ = TaskQ(handler=handler).create_job(name="job2")

    in_task1 = Task(entrypoint=dummy_args_task, level=2, name="taska")
    in_task2 = Task(entrypoint=dummy_args_task, level=1, name="taskb")
    in_task3 = Task(entrypoint=dummy_args_task, level=1, name="taskb")
    in_task4 = Task(entrypoint=dummy_args_task, level=1, name="taskd")
    in_task5 = Task(entrypoint=dummy_args_task, level=2, name="taske")

    taskq1.add_tasks(
        [
            in_task2,
            in_task1,
            in_task3,
        ]
    )

    taskq2.add_tasks(
        [
            in_task5,
            in_task4,
        ]
    )

    jid1 = taskq1.job_id
    jid2 = taskq2.job_id
    tids = []

    # sanity check
    assert len(Job.get_all(_handler=handler)) == 2
    assert len(taskq1.get_tasks()) == 3
    assert len(Job.get_all(_handler=handler)) == 2
    assert len(taskq2.get_tasks()) == 2

    # taskq 1
    action, task = taskq1._take_next_task(level=None)
    assert action == EAction.RUN_TASK
    assert task.task_id not in tids
    tids.append(task.task_id)
    _compare_tasks(in_task2, task, job_id=jid1)
    tids.append(task.task_id)

    action, task = taskq1._take_next_task(level=None)
    assert action == EAction.RUN_TASK
    assert task.task_id not in tids
    tids.append(task.task_id)
    _compare_tasks(in_task3, task, job_id=jid1)  # note in_task3 is copy of in_task1

    action, task = taskq1._take_next_task(level=None)
    assert action == EAction.WAIT
    assert task is None

    # taskq 2
    action, task = taskq2._take_next_task(level=None)
    assert action == EAction.RUN_TASK
    assert task.task_id not in tids
    tids.append(task.task_id)
    _compare_tasks(in_task4, task, job_id=jid2)
    tids.append(task.task_id)

    action, task = taskq2._take_next_task(level=None)
    assert action == EAction.WAIT
    assert task is None


def test_take_next_all_jobs(config):
    handler = from_config(config)
    taskq1: TaskQ = TaskQ(handler=handler).create_job(name="job1")
    taskq2: TaskQ = TaskQ(handler=handler).create_job(name="job2")

    def init_task(name):
        return Task(entrypoint=dummy_args_task, name=name)

    in_task1 = init_task("job1 - task1")
    in_task2 = init_task("job1 - task2")
    taskq1.add_tasks(
        [
            in_task1,
            in_task2,
        ]
    )

    in_task4 = init_task("job2 - task4")
    in_task5 = init_task("job2 - task5")
    taskq2.add_tasks(
        [
            in_task4,
            in_task5,
        ]
    )

    in_task3 = init_task("job1 - task3")
    taskq1.add_tasks(
        [
            in_task3,
        ]
    )

    jid1 = taskq1.job_id
    jid2 = taskq2.job_id

    # sanity check
    assert len(Job.get_all(_handler=handler)) == 2
    assert len(taskq1.get_tasks()) == 3
    assert len(Job.get_all(_handler=handler)) == 2
    assert len(taskq2.get_tasks()) == 2

    taskq = TaskQ(config=config)
    in_tasks = [in_task1, in_task2, in_task3, in_task4, in_task5]
    in_jids = [jid1, jid1, jid1, jid2, jid2]
    for t, jid in zip(in_tasks, in_jids):
        action, task = taskq._take_next_task()
        assert action == EAction.RUN_TASK
        assert task.task_id == t.task_id
        _compare_tasks(t, task, job_id=jid)

    assert not monotonic([t.task_id for t in in_tasks])


def test_run_default(config, tmp_path: Path):
    filepath = tmp_path / "file.txt"

    taskq = TaskQ(config=config).create_job()

    taskq.add_tasks(
        [
            Task(entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file", targs=targs(filepath, "task 0\n")),
            Task(entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file", targs=targs(filepath, "task 1\n")),
            Task(entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file", targs=targs(filepath, "task 2\n")),
        ]
    )

    taskq.run()

    assert filepath.exists()
    assert filepath.read_text() == "task 0\n" "task 1\n" "task 2\n"


def test_run_task_raise_exception(config):
    # no exception raised
    try:
        config["run"]["raise_exception"] = False
        taskq: TaskQ = TaskQ(config=config).create_job()
        taskq.add_tasks(
            [
                Task(entrypoint="ataskq.tasks_utils.exception_task", targs=targs(message="task failed")),
            ]
        )
        taskq.run()
    except Exception:
        assert False, "exception_task raises exception with run_task_raise_exception=False"

    # exception raised
    config["run"]["raise_exception"] = True
    taskq: TaskQ = TaskQ(config=config).create_job()
    taskq.add_tasks(
        [
            Task(entrypoint="ataskq.tasks_utils.exception_task", targs=targs(message="task failed")),
        ]
    )
    with pytest.raises(Exception) as excinfo:
        taskq.run()
    assert excinfo.value.args[0] == "task failed"


def test_run_2_processes(config, tmp_path: Path):
    filepath = tmp_path / "file.txt"

    taskq = TaskQ(config=config).create_job()

    taskq.add_tasks(
        [
            Task(
                entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file_mp_lock",
                targs=targs(filepath, "task 0\n"),
            ),
            Task(
                entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file_mp_lock",
                targs=targs(filepath, "task 1\n"),
            ),
            Task(
                entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file_mp_lock",
                targs=targs(filepath, "task 2\n"),
            ),
        ]
    )

    taskq.run(concurrency=2)

    assert filepath.exists()
    text = filepath.read_text()
    assert "task 0\n" in text
    assert "task 1\n" in text
    assert "task 1\n" in text


def test_run_all_jobs(config, tmp_path: Path):
    filepath1 = tmp_path / "file1.txt"
    taskq1 = TaskQ(config=config).create_job()
    taskq1.add_tasks(
        [
            Task(
                entrypoint=write_to_file,
                targs=targs(filepath1, "task 0\n"),
            ),
        ]
    )

    filepath2 = tmp_path / "file2.txt"
    taskq2 = TaskQ(config=config).create_job()
    taskq2.add_tasks(
        [
            Task(
                entrypoint=write_to_file,
                targs=targs(filepath2, "task 1\n"),
            ),
        ]
    )

    TaskQ(config=config).run()

    assert filepath1.exists()
    assert "task 0\n" == filepath1.read_text()
    assert filepath2.exists()
    assert "task 1\n" == filepath2.read_text()


@pytest.mark.parametrize("num_processes", [None, 2])
def test_run_by_level(config, tmp_path: Path, num_processes: int):
    filepath = tmp_path / "file.txt"

    taskq = TaskQ(config=config).create_job()

    taskq.add_tasks(
        [
            Task(
                level=0,
                entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file_mp_lock",
                targs=targs(filepath, "task 0\n"),
            ),
            Task(
                level=1,
                entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file_mp_lock",
                targs=targs(filepath, "task 1\n"),
            ),
            Task(
                level=1,
                entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file_mp_lock",
                targs=targs(filepath, "task 2\n"),
            ),
            Task(
                level=2,
                entrypoint="ataskq.tasks_utils.write_to_file_tasks.write_to_file_mp_lock",
                targs=targs(filepath, "task 3\n"),
            ),
        ]
    )

    assert taskq.count_pending_tasks_below_level(3) == 4

    assert taskq.count_pending_tasks_below_level(1) == 1
    taskq.run(level=0, concurrency=num_processes)
    taskq.count_pending_tasks_below_level(1) == 0
    assert filepath.exists()
    text = filepath.read_text()
    assert "task 0\n" in text

    assert taskq.count_pending_tasks_below_level(2) == 2
    taskq.run(level=1, concurrency=num_processes)
    taskq.count_pending_tasks_below_level(2) == 0
    text = filepath.read_text()
    assert "task 0\n" in text
    assert "task 1\n" in text
    assert "task 2\n" in text

    assert taskq.count_pending_tasks_below_level(3) == 1
    taskq.run(level=2, concurrency=num_processes)
    taskq.count_pending_tasks_below_level(3) == 0
    text = filepath.read_text()
    assert "task 0\n" in text
    assert "task 1\n" in text
    assert "task 2\n" in text
    assert "task 3\n" in text


def test_monitor_pulse_failure(config):
    # set monitor pulse longer than timeout
    config["monitor"]["pulse_interval"] = 2
    config["monitor"]["pulse_timeout"] = 1.5
    taskq = TaskQ(config=config).create_job()
    taskq.add_tasks(
        [
            # reserved keyward for ignored task for testing
            Task(entrypoint="ataskq.skip_run_task", targs=targs("task will fail")),
            Task(entrypoint=dummy_args_task, targs=targs("task will success")),
        ]
    )
    start = datetime.now()
    taskq.run()
    stop = datetime.now()

    tasks = taskq.get_tasks()

    assert tasks[0].status == EStatus.FAILURE
    assert tasks[1].status == EStatus.SUCCESS
    assert stop - start > timedelta(seconds=1.5)


def test_task_wait_timeout(config):
    # set monitor pulse longer than timeout
    config["run"]["raise_exception"] = True
    config["run"]["wait_timeout"] = 0
    taskq = TaskQ(config=config).create_job()
    taskq.add_tasks(
        [
            Task(entrypoint=dummy_args_task, level=1, targs=targs("task will success", sleep=0.2)),
            Task(entrypoint=dummy_args_task, level=2, targs=targs("task will success")),
        ]
    )

    with pytest.raises(Exception) as excinfo:
        taskq.run(concurrency=2)
    assert excinfo.value.args[0] == "Some processes failed, see logs for details"


def test_max_jobs(config):
    max_jobs = 10
    config["db"]["max_jobs"] = max_jobs
    taskq = TaskQ(config=config)
    if not isinstance(taskq.handler, DBHandler):
        pytest.skip()

    jobs_id = []
    for i in range(max_jobs * 2):
        taskq.clear_job()
        taskq.create_job(name=f"job{i}")
        jobs_id.append(taskq.job.job_id)
    jobs = Job.get_all(taskq.handler)
    assert len(jobs) == 10

    remaining_jobs = [j.job_id for j in jobs]
    assert remaining_jobs == jobs_id[-max_jobs:]


def test_run_forever(config):
    taskq = TaskQ(config=config)
    taskq.run()  # no tasks, instant complete

    def run():
        TaskQ(config=[config, {"run": {"run_forever": True}}]).run()

    p = Process(target=run)
    p.start()
    while not p.is_alive():
        time.sleep(0.2)
    time.sleep(2)
    assert p.is_alive(), "run finished with run_forever True"
    p.kill()
    p.join()



from typing import Union, List, Dict
from enum import Enum
import pickle
from importlib import import_module
from datetime import datetime
from copy import copy

from .imodel import IModel, IModelSerializer
from .handler import get_handler, Handler


class EntryPointRuntimeError(RuntimeError):
    pass


class TARGSLoadRuntimeError(EntryPointRuntimeError):
    pass


class EntrypointLoadRuntimeError(EntryPointRuntimeError):
    pass


class EntrypointCallRuntimeError(EntryPointRuntimeError):
    pass


class EStatus(str, Enum):
    PENDING = "pending"
    RUNNING = "running"
    SUCCESS = "success"
    FAILURE = "failure"

    def __str__(self) -> str:
        return self.value


class EntryPoint:
    @staticmethod
    def init(kwargs) -> None:
        entrypoint = kwargs.get("entrypoint")
        if callable(entrypoint):
            kwargs["entrypoint"] = f"{entrypoint.__module__}.{entrypoint.__name__}"

        targs = kwargs.get("targs")
        if targs is not None and isinstance(targs, tuple):
            assert len(targs) == 2
            assert isinstance(targs[0], tuple)
            assert isinstance(targs[1], dict)
            kwargs["targs"] = pickle.dumps(targs)

    def get_targs(self):
        if self.targs is not None:
            try:
                targs = pickle.loads(self.targs)
                assert len(targs) == 2, "targs must be tuple of 2 elements"
                assert isinstance(targs[0], tuple), "targs[0] must be args tuple"
                assert isinstance(targs[1], dict), "targs[0] must be kwargs dict"
            except Exception as ex:
                raise TARGSLoadRuntimeError() from ex

        else:
            targs = ((), {})

        return targs[0], targs[1]

    def get_entrypoint(self):
        ep = self.entrypoint

        try:
            assert "." in ep, "entry point must be inside a module."
            module_name, func_name = ep.rsplit(".", 1)
            m = import_module(module_name)
            assert hasattr(
                m, func_name
            ), f"failed to load entry point, module '{module_name}' doen't have func named '{func_name}'."
            func = getattr(m, func_name)
            assert callable(func), f"entry point is not callable, '{module_name}.{func}'."
        except ImportError as ex:
            raise EntrypointLoadRuntimeError(f"Failed to load module '{module_name}'. Exception: '{ex}'")
        except Exception as ex:
            raise EntrypointLoadRuntimeError(f"Failed to load entry point '{ep}'. Exception: '{ex}'") from ex

        return func

    def call(self):
        args, kwargs = self.get_targs()
        entrypoint = self.get_entrypoint()

        try:
            ret = entrypoint(*args, **kwargs)
        except Exception as ex:
            raise EntrypointCallRuntimeError(
                f"Failed while call entrypoint function '{self.entrypoint}'. Exception: '{ex}'"
            ) from ex

        return ret


def _handle_union(cls_name, member, annotations, value, type_handlers=None):
    if type_handlers is None:
        type_handlers = dict()

    # check if value is of supported types
    for ann in annotations:
        if isinstance(value, ann):
            return value

    # attemp cast value
    success = False
    value = None
    for ann in annotations:
        try:
            if ann in type_handlers:
                value = type_handlers[ann](value)
            else:
                value = ann(value)
            success = True
            break
        except Exception:
            continue

    if not success:
        raise Exception(f"{cls_name}::{member}({annotations}) failed casting {type(value)} - '{value}'.")

    return value


class Model(IModel):
    def __init__(self, _serialize=True, **kwargs) -> None:
        cls_annotations = self.__annotations__
        defaults = getattr(self, "__DEFAULTS__", dict())

        # check a kwargs are class members
        for k in kwargs.keys():
            if k not in cls_annotations.keys():
                raise Exception(f"'{k}' not a possible class '{self.__class__.__name__}' member.")

        # set defaults
        for member in cls_annotations.keys():
            # default None to members not passed
            if member not in kwargs:
                kwargs[member] = defaults.get(member)

        # annotate kwargs
        if _serialize:
            kwargs = self._serialize(kwargs, dict())  # flag passed on constructor with no interface handlers

        # set kwargs as class members
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def _serialize(cls, kwargs: dict, type_handlers: dict):
        ret = dict()
        cls_annotations = cls.__annotations__
        cls_name = cls.__name__
        for k, v in kwargs.items():
            if k not in cls_annotations:
                raise Exception(f"interface key '{k}' not in model annotations.")

            # allow None values (no handling)
            if v is None:
                ret[k] = None
                continue

            # get member annotation
            annotation = cls_annotations[k]

            # handle union
            if getattr(annotation, "__origin__", None) is Union:
                ret[k] = _handle_union(cls_name, k, annotation.__args__, v, type_handlers)
                continue

            # Single annotation cast
            ann = cls_annotations[k]

            ann_name = None
            if ann in type_handlers:
                ann_name = f"type_handler[{ann.__name__}]"
                ann = type_handlers[ann]
            elif issubclass(ann, str) and str in type_handlers:
                # string subclasses
                ann_name = f"type_handler[{ann.__name__} - str sublcass]"
                ann = type_handlers[str]
            elif issubclass(ann, Enum) and Enum in type_handlers:
                # string subclasses
                ann_name = f"type_handler[{ann.__name__} - Enum sublcass]"
                ann = type_handlers[Enum]
            else:
                # check if already in relevant type
                if isinstance(v, ann):
                    ret[k] = v
                    continue

                ann_name = f"{ann.__name__}"
                ann = ann

            try:
                ret[k] = ann(v)
            except Exception as ex:
                raise Exception(f"{cls_name}::{k}({ann_name}) failed cast from '{v}'({type(v).__name__})") from ex

        return ret

    @classmethod
    def i2m(cls, kwargs: Union[dict, List[dict]], serializer: IModelSerializer) -> Union[dict, List[dict]]:
        """interface to model"""
        if isinstance(kwargs, list):
            ret = [cls._serialize(kw, serializer.i2m_serialize()) for kw in kwargs]
        else:
            ret = cls._serialize(kwargs, serializer.i2m_serialize())

        return ret

    @classmethod
    def from_interface(cls, kwargs: Union[dict, List[dict]], serializer: IModelSerializer):
        """interface to model"""
        mkwargs = cls.i2m(kwargs, serializer)
        if isinstance(kwargs, list):
            ret = [cls(_serialize=False, **kw) for kw in mkwargs]
        else:
            ret = cls(_serialize=False, **mkwargs)

        return ret

    @classmethod
    def m2i(cls, kwargs: Union[dict, List[dict]], serializer: IModelSerializer) -> Union[dict, List[dict]]:
        """model to interface"""
        if isinstance(kwargs, list):
            ret = [cls._serialize(kw, serializer.m2i_serialize()) for kw in kwargs]
        else:
            ret = cls._serialize(kwargs, serializer.m2i_serialize())

        return ret

    def to_interface(self, serializer: IModelSerializer) -> dict:
        """model to interface"""
        ret = self.m2i(self.__dict__, serializer)

        return ret

    @classmethod
    def count_all(cls, _handler: Handler = None, **kwargs):
        if _handler is None:
            _handler = get_handler(assert_registered=True)

        ret = _handler.count_all(cls, **kwargs)
        return ret

    @classmethod
    def get_all_dict(cls, _handler: Handler = None, **kwargs):
        if _handler is None:
            _handler = get_handler(assert_registered=True)

        ret = _handler.get_all(cls, **kwargs)
        ret = cls.i2m(ret, _handler)

        return ret

    @classmethod
    def get_all(cls, _handler: Handler = None, **kwargs):
        ret = cls.get_all_dict(_handler=_handler, **kwargs)
        ret = [cls(**r, _serialize=False) for r in ret]

        return ret

    @classmethod
    def get_dict(cls, model_id: int, _handler: Handler = None):
        if _handler is None:
            _handler = get_handler(assert_registered=True)

        ikwargs = _handler.get(cls, model_id)
        mkwargs = cls.i2m(ikwargs, _handler)

        return mkwargs

    @classmethod
    def get(cls, model_id: int, _handler: Handler = None):
        mkwargs = cls.get_dict(model_id, _handler)
        ret = cls(**mkwargs, _serialize=False)

        return ret

    def create(self, _handler: Handler = None, **mkwargs):
        if not mkwargs:
            assert (
                getattr(self, self.id_key()) is None
            ), f"id '{self.id_key()}' can't be assigned when creating '{self.__class__.__name__}({self.table_key()})'"
            mkwargs = copy(self.__dict__)
            mkwargs.pop(self.id_key())

        assert (
            self.id_key() not in mkwargs
        ), f"id '{self.id_key()}' can't be passed to create '{self.__class__.__name__}({self.table_key()})'"

        if _handler is None:
            _handler = get_handler(assert_registered=True)

        ikwargs = self.m2i(mkwargs, _handler)
        model_id = _handler._create(self.__class__, **ikwargs)

        setattr(self, self.id_key(), model_id)

        return self

    def update(self, _handler: Handler = None, **mkwargs):
        if not mkwargs:
            assert (
                getattr(self, self.id_key()) is not None
            ), f"id '{self.id_key()}' must be assigned when updating '{self.__class__.__name__}({self.table_key()})'"
            mkwargs = copy(self.__dict__)
            mkwargs.pop(self.id_key())

        assert (
            self.id_key() not in mkwargs
        ), f"id '{self.id_key()}' can't be passed to update '{self.__class__.__name__}({self.table_key()})'"

        if _handler is None:
            _handler = get_handler(assert_registered=True)

        model_id = getattr(self, self.id_key())
        ikwargs = self.m2i(mkwargs, _handler)
        _handler._update(self.__class__, model_id, **ikwargs)

        for k, v in mkwargs.items():
            setattr(self, k, v)

        return self

    @classmethod
    def delete_all(cls, _handler: Handler = None, **kwargs):
        if _handler is None:
            _handler = get_handler(assert_registered=True)

        ret = _handler.delete_all(cls, **kwargs)
        return ret

    def delete(self, _handler: Handler = None):
        model_id = getattr(self, self.id_key())
        assert (
            model_id is not None
        ), f"id '{self.__class__.__name__}({self.table_key()} -> {self.id_key()})' required for delete"

        if _handler is None:
            _handler = get_handler(assert_registered=True)

        _handler.delete(self.__class__, model_id)
        setattr(self, self.id_key(), None)

        return self

    def add_children(
        self,
        child_cls: IModel,
        children: Union[Union[IModel, dict], List[Union[IModel, dict]]],
        _handler: Handler = None,
    ):
        if not children:
            return

        if not isinstance(children, list):
            children = [children]

        assert child_cls in self.children(), f"no children association defined for '{child_cls}'"
        parent_key = self.children()[child_cls]
        parent_key_val = getattr(self, self.id_key())

        children_mkwargs = []
        for i, c in enumerate(children):
            if isinstance(c, child_cls):
                assert (
                    getattr(c, c.id_key()) is None
                ), f"id '{child_cls.id_key()}' can't be assigned when creating '{child_cls.__name__}({child_cls.table_key()})'"
                mkwargs = copy(c.__dict__)
                mkwargs.pop(c.id_key())
            elif isinstance(c, dict):
                mkwargs = c
            else:
                raise Exception(f"item [{i}]: Unsupported child type '{type(c)}'")

            # assign parent key to child
            assert (
                mkwargs.get(parent_key) is None
            ), f"child '{child_cls.__name__}[{i}]' can't have parent key '{self.__class__.__name__}->{parent_key}' assigned"
            mkwargs[parent_key] = parent_key_val
            children_mkwargs.append(mkwargs)

        if _handler is None:
            _handler = get_handler(assert_registered=True)

        child_ids = _handler.create_bulk(child_cls, children_mkwargs)
        for cid, c in zip(child_ids, children):
            setattr(c, parent_key, parent_key_val)
            setattr(c, child_cls.id_key(), cid)

        return children

    def get_children_dict(self, child_cls: IModel, _handler: Handler = None):
        assert child_cls in self.children(), f"no children association defined for '{child_cls}'"
        parent_key = self.children()[child_cls]
        primary_key_val = getattr(self, self.id_key())

        if _handler is None:
            _handler = get_handler(assert_registered=True)

        ikwargs = _handler.get_all(child_cls, **{f"{parent_key}": primary_key_val})
        mkwargs = child_cls.i2m(ikwargs, _handler)

        return mkwargs

    def get_children(self, child_cls: IModel, _handler: Handler = None):
        mkwargs = self.get_children_dict(child_cls, _handler)
        ret = [child_cls(**kw, _serialize=False) for kw in mkwargs]

        return ret


class Task(Model, EntryPoint):
    task_id: int
    name: str
    level: float
    entrypoint: str
    targs: bytes
    status: EStatus
    take_time: datetime
    start_time: datetime
    done_time: datetime
    pulse_time: datetime
    description: str
    # summary_cookie = None,
    job_id: int

    __DEFAULTS__ = dict(status=EStatus.PENDING, entrypoint="", level=0.0)

    @staticmethod
    def id_key():
        return "task_id"

    @staticmethod
    def table_key():
        return "tasks"

    def __init__(self, **kwargs) -> None:
        EntryPoint.init(kwargs)
        Model.__init__(self, **kwargs)

    def __str__(self):
        return f"{self.name}({self.task_id})" if self.name else f"{self.task_id}"


class Job(Model):
    job_id: int
    name: str
    priority: float
    description: str

    @staticmethod
    def id_key():
        return "job_id"

    @staticmethod
    def table_key():
        return "jobs"

    @staticmethod
    def children():
        return {
            Task: "job_id",
        }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_tasks(self, _handler=None) -> List[Task]:
        return self.get_children(Task, _handler=_handler)

    def add_tasks(self, tasks: List[Task], _handler=None):
        return self.add_children(Task, tasks, _handler=_handler)


__MODELS__: Dict[str, Model] = {m.table_key(): m for m in [Task, Job]}



import pytest

from . import TaskQ, Task
from .handler import from_config


@pytest.fixture()
def job_ids(config):
    ret = []

    taskq = TaskQ(config=config).create_job(name="job1")
    taskq.add_tasks(
        [
            Task(name="n1", entrypoint=""),
            Task(name="n1", entrypoint=""),
            Task(name="n2", entrypoint=""),
        ]
    )
    ret.append(taskq.job_id)

    taskq = TaskQ(config=config).create_job(name="job2")
    taskq.add_tasks(
        [
            Task(name="n3", entrypoint=""),
            Task(name="n4", entrypoint=""),
            Task(name="n4", entrypoint=""),
        ]
    )
    ret.append(taskq.job_id)

    return ret


@pytest.fixture()
def handler(config):
    return from_config(config)


def test_tasks_status(handler, job_ids):
    # job 1
    status = handler.tasks_status(job_id=job_ids[0])

    assert len(status) == 2

    assert status[0]["total"] == 2
    assert status[0]["name"] == "n1"
    assert status[0]["pending"] == 2

    assert status[1]["total"] == 1
    assert status[1]["name"] == "n2"
    assert status[1]["pending"] == 1

    # job 2
    status = handler.tasks_status(job_id=job_ids[1])

    assert len(status) == 2

    assert status[0]["total"] == 1
    assert status[0]["name"] == "n3"
    assert status[0]["pending"] == 1

    assert status[1]["total"] == 2
    assert status[1]["name"] == "n4"
    assert status[1]["pending"] == 2


def test_jobs_status(handler, job_ids):
    # job 1
    status = handler.jobs_status()

    assert len(status) == 2

    # order desc
    assert status[1]["job_id"] == job_ids[0]
    assert status[1]["name"] == "job1"
    assert status[1]["tasks"] == 3
    assert status[1]["pending"] == 3

    assert status[0]["job_id"] == job_ids[1]
    assert status[0]["name"] == "job2"
    assert status[0]["tasks"] == 3
    assert status[0]["pending"] == 3



from typing import List, NamedTuple, Tuple, Union
from enum import Enum
from datetime import datetime
import base64

from ataskq.imodel import IModel

try:
    import requests
except ImportError:
    raise Exception("'requests' is required to use ataskq REST handler.")

from .handler import Handler, EAction, from_datetime, get_query_kwargs, to_datetime


class RESTConnection(NamedTuple):
    url: Union[None, str]

    def __str__(self):
        return {self.url}


class RESTHandler(Handler):
    # todo: remove max jobs
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    @staticmethod
    def from_connection_str(conn):
        ret = RESTConnection(url=conn)

        return ret

    @staticmethod
    def m2i_serialize():
        type_handlers = {
            datetime: lambda v: from_datetime(v),
            Enum: lambda v: v.value,
            bytes: lambda v: base64.b64encode(v).decode("ascii"),
        }

        return type_handlers

    @staticmethod
    def i2m_serialize():
        type_handlers = {datetime: lambda v: to_datetime(v), bytes: lambda v: base64.b64decode(v.encode("ascii"))}

        return type_handlers

    @property
    def api_url(self):
        return f"{self._connection.url}/api"

    def rest_get(self, url, *args, **kwargs):
        url = f"{self.api_url}/{url}"
        res = requests.get(url, *args, **kwargs)
        assert res.ok, f"get url '{url}' failed. message: {res.text}"

        return res.json()

    def rest_post(self, url, *args, **kwargs):
        url = f"{self.api_url}/{url}"
        res = requests.post(url, *args, **kwargs)
        assert res.ok, f"post url '{url}' failed. message: {res.text}"

        return res.json()

    def rest_put(self, url, *args, **kwargs):
        url = f"{self.api_url}/{url}"
        res = requests.put(url, *args, **kwargs)
        assert res.ok, f"put url '{url}' failed. message: {res.text}"

        return res.json()

    def rest_delete(self, url, *args, **kwargs):
        url = f"{self.api_url}/{url}"
        res = requests.delete(url, *args, **kwargs)
        assert res.ok, f"delete url '{url}' failed. message: {res.text}"

        return res.json()

    #########
    # Model #
    #########
    def get_all(self, model_cls: IModel, **kwargs) -> List[dict]:
        query_kwargs = get_query_kwargs(kwargs)
        res = self.rest_get(model_cls.table_key(), params=query_kwargs)
        return res

    def get(self, model_cls: IModel, model_id: int) -> dict:
        res = self.rest_get(f"{model_cls.table_key()}/{model_id}")
        return res

    def count_all(self, model_cls: IModel, **kwargs) -> int:
        query_kwargs = get_query_kwargs(kwargs)
        res = self.rest_get(f"{model_cls.table_key()}/count", params=query_kwargs)
        return res

    def _create(self, model_cls: IModel, **ikwargs: dict) -> int:
        res = self.rest_post(model_cls.table_key(), json=ikwargs)

        return res

    def _create_bulk(self, model_cls: IModel, ikwargs: List[dict]) -> List[int]:
        res = self.rest_post(f"{model_cls.table_key()}/bulk", json=ikwargs)

        return res

    def delete_all(self, model_cls: IModel, **kwargs):
        query_kwargs = get_query_kwargs(kwargs)
        self.rest_delete(f"{model_cls.table_key()}", json=query_kwargs)

    def delete(self, model_cls: IModel, model_id: int):
        self.rest_delete(f"{model_cls.table_key()}/{model_id}")

    def _update(self, model_cls: IModel, model_id, **ikwargs):
        self.rest_put(f"{model_cls.table_key()}/{model_id}", json=ikwargs)

    def update_all(self, model_cls: IModel, **ikwargs):
        self.rest_put(f"{model_cls.table_key()}", json=ikwargs)

    ##################
    # Custom Queries #
    ##################

    def take_next_task(self, **kwargs) -> Tuple:
        from ..models import Task

        res = self.rest_get("custom_query/take_next_task", params=kwargs)

        action = EAction(res["action"])
        task = self.from_interface(Task, res["task"]) if res["task"] is not None else None

        return (action, task)

    def tasks_status(self, **kwargs):
        res = self.rest_get(f"custom_query/tasks_status", params=kwargs)

        return res

    def jobs_status(self, **kwargs):
        res = self.rest_get(f"custom_query/jobs_status", params=kwargs)

        return res



# todo: add queries order_by tests

import pytest

from .config import load_config
from .handler import Handler, from_config
from .handler.db_handler import DBHandler, transaction_decorator
from .handler import register_handler


@pytest.fixture
def handler(config) -> Handler:
    handler = from_config(config)
    register_handler("test_handler", handler)
    return handler


class ForTestDBHandler:
    def __init__(self, handler):
        self._handler = handler
        self.error_msg = []

    def connect(self):
        return self._handler.connect()

    def __getattr__(self, name):
        return getattr(self._handler, name)

    @transaction_decorator()
    def invalid(self, c):
        c.execute("SOME INVALID TRANSACTION")


def test_invalid_transaction(handler):
    if not isinstance(handler, DBHandler):
        pytest.skip()

    handler = ForTestDBHandler(handler)
    with pytest.raises(Exception) as excinfo:
        handler.invalid()
    assert "syntax error" in str(excinfo.value)


def test_db_format(config, handler):
    assert isinstance(handler, Handler)
    conn = config["connection"]

    if "sqlite" in conn:
        from .handler.sqlite3 import SQLite3DBHandler

        assert isinstance(handler, SQLite3DBHandler)
        assert "ataskq.db" in handler.db_path
    elif "pg" in conn:
        from .handler.postgresql import PostgresqlDBHandler

        assert isinstance(handler, PostgresqlDBHandler)
    elif "http" in conn:
        from .handler.rest_handler import RESTHandler

        assert isinstance(handler, RESTHandler)
    else:
        raise Exception(f"unknown handler type in connection string '{conn}'")


def test_db_invalid_format_no_sep():
    with pytest.raises(RuntimeError) as excinfo:
        from_config(config=load_config({"connection": "sqlite"}, environ=False))
    assert "connection must be of format <type>://<connection string>" == str(excinfo.value)


def test_db_invalid_format_no_type():
    with pytest.raises(RuntimeError) as excinfo:
        from_config(load_config({"connection": f"://ataskq.db"}, environ=False))
    assert "missing handler type, connection must be of format <type>://<connection string>" == str(excinfo.value)


def test_db_invalid_format_no_connectino():
    with pytest.raises(RuntimeError) as excinfo:
        from_config(load_config({"connection": "sqlite://"}, environ=False))
    assert "missing connection string, connection must be of format <type>://<connection string>" == str(excinfo.value)



import re
from typing import NamedTuple, Union
from datetime import datetime

try:
    import psycopg2
except ModuleNotFoundError:
    raise Exception("'psycopg2' is reuiqred for using atasgq postgresql adapter.")

from .db_handler import DBHandler
from .handler import to_datetime, from_datetime


class PostgresConnection(NamedTuple):
    user: Union[None, str]
    password: Union[None, str]
    host: str
    port: int
    database: str

    def __str__(self):
        if self.user:
            userspec = f"{self.user}" + (self.password and f":{self.password}") + "@"

        return f"pg://{userspec}{self.host}:{self.port}/{self.database}"


class PostgresqlDBHandler(DBHandler):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    @staticmethod
    def from_connection_str(conn):
        # https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-CONNSTRING-URIS
        # todo: add params spec support
        format = "pg://[user[:password]@][host][:port][/database]"
        pattern = r"pg://(?P<user>[^:@]+)(:(?P<password>[^@]+))?@?(?P<host>[^:/]+)(:(?P<port>\d+))?/(?P<database>.+)$"

        match = re.match(pattern, conn)

        if not match:
            raise Exception(f"db must be in '{format}', ex: 'pg://user:password@localhost:5432/mydb'")

        user = match.group("user")
        password = match.group("password")
        host = match.group("host")
        port = match.group("port")
        database = match.group("database")
        ret = PostgresConnection(user=user, password=password, host=host, port=port, database=database)

        return ret

    @staticmethod
    def m2i_serialize():
        type_handlers = {
            datetime: lambda v: from_datetime(v),
        }

        return type_handlers

    @staticmethod
    def i2m_serialize():
        type_handlers = {
            datetime: lambda v: to_datetime(v),
        }

        return type_handlers

    @property
    def format_symbol(self):
        return "%s"

    @property
    def connection(self):
        return self._connection

    @property
    def bytes_type(self):
        return "BYTEA"

    @property
    def primary_key(self):
        return "SERIAL PRIMARY KEY"

    @property
    def timestamp_type(self):
        return "TIMESTAMP"

    def timestamp(self, ts):
        return f"'{ts}'::timestamp"

    @property
    def begin_exclusive(self):
        return "BEGIN"

    @property
    def for_update(self):
        return "FOR UPDATE"

    def connect(self):
        conn = psycopg2.connect(
            host=self.connection.host,
            database=self.connection.database,
            user=self.connection.user,
            password=self.connection.password,
        )
        return conn
